from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0235(RuleTemplate):
    """
    CHK0235: A masking is not expected to be defined for any study role in a study design with an open label blinding schema.

    Applies to: StudyRole
    Attributes: masking
    """

    def __init__(self):
        super().__init__(
            "CHK0235",
            RuleTemplate.ERROR,
            "A masking is not expected to be defined for any study role in a study design with an open label blinding schema.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .administrable_product_property import AdministrableProductProperty
from .identifier import AdministrableProductIdentifier
from .ingredient import Ingredient
from .code import Code
from .alias_code import AliasCode
from .comment_annotation import CommentAnnotation


class AdministrableProduct(ApiBaseModelWithIdNameLabelAndDesc):
    pharmacologicClass: Union[Code, None] = None
    administrableDoseForm: AliasCode
    productDesignation: Code
    sourcing: Union[Code, None] = None
    properties: List[AdministrableProductProperty] = []
    identifiers: List[AdministrableProductIdentifier] = []
    ingredients: List[Ingredient] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["AdministrableProduct"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0247(RuleTemplate):
    """
    CHK0247: The sponsor study role must be applicable to a study version.

    Applies to: StudyRole
    Attributes: appliesTo
    """

    def __init__(self):
        super().__init__(
            "CHK0247",
            RuleTemplate.ERROR,
            "The sponsor study role must be applicable to a study version.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0188(RuleTemplate):
    """
    CHK0188: A piece of narrative content must only reference narrative content items that have been defined within the study version as the narrative content.

    Applies to: NarrativeContent
    Attributes: contentItem
    """

    def __init__(self):
        super().__init__(
            "CHK0188",
            RuleTemplate.ERROR,
            "A piece of narrative content must only reference narrative content items that have been defined within the study version as the narrative content.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .schedule_timeline_exit import ScheduleTimelineExit
from .scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from .timing import Timing
from .duration import Duration


class ScheduleTimeline(ApiBaseModelWithIdNameLabelAndDesc):
    mainTimeline: bool
    entryCondition: str
    entryId: str
    exits: List[ScheduleTimelineExit] = []
    timings: List[Timing] = []
    instances: List[Union[ScheduledDecisionInstance, ScheduledActivityInstance]] = []
    plannedDuration: Union[Duration, None] = None
    instanceType: Literal["ScheduleTimeline"]

    def first_timepoint(self) -> ScheduledActivityInstance:
        return self.instances[0] if self.instances else None

    def find_timepoint(
        self, id: str
    ) -> ScheduledActivityInstance | ScheduledDecisionInstance:
        return next((x for x in self.instances if x.id == id), None)

    def find_exit(self, id: str) -> ScheduleTimelineExit:
        return next((x for x in self.exits if x.id == id), None)

    def timepoint_list(self) -> list:
        return self.instances

    def find_timing_from(self, id: str) -> Timing:
        return next(
            (x for x in self.timings if x.relativeFromScheduledInstanceId == id), None
        )

    def find_timing_to(self, id: str) -> Timing:
        return next(
            (x for x in self.timings if x.relativeToScheduledInstanceId == id), None
        )



from datetime import date
from typing import List, Literal, Union
from typing_extensions import deprecated
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .identifier import StudyIdentifier, ReferenceIdentifier
from .study_design import (
    StudyDesign,
    InterventionalStudyDesign,
    ObservationalStudyDesign,
)
from .governance_date import GovernanceDate
from .study_amendment import StudyAmendment
from .study_title import StudyTitle
from .eligibility_criterion import EligibilityCriterionItem
from .narrative_content import NarrativeContentItem
from .comment_annotation import CommentAnnotation
from .abbreviation import Abbreviation
from .study_role import StudyRole
from .organization import Organization
from .study_intervention import StudyIntervention
from .administrable_product import AdministrableProduct
from .medical_device import MedicalDevice
from .product_organization_role import ProductOrganizationRole
from .biomedical_concept import BiomedicalConcept
from .biomedical_concept_category import BiomedicalConceptCategory
from .biomedical_concept_surrogate import BiomedicalConceptSurrogate
from .syntax_template_dictionary import SyntaxTemplateDictionary
from .study_definition_document import (
    StudyDefinitionDocument,
    StudyDefinitionDocumentVersion,
)
from .condition import Condition
from .extension import Extension


CS_EXT_URL = "www.d4k.dk/usdm/extensions/001"  # Confidentiality statement
OV_EXT_URL = "www.d4k.dk/usdm/extensions/002"  # Original protocol (original version)
CC_EXT_URL = "www.d4k.dk/usdm/extensions/004"  # Compund codes
CN_EXT_URL = "www.d4k.dk/usdm/extensions/005"  # Compund names
ME_EXT_URL = "www.d4k.dk/usdm/extensions/006"  # Medical Expert
SS_EXT_URL = "www.d4k.dk/usdm/extensions/007"  # Sponsor signatory
SAL_EXT_URL = "www.d4k.dk/usdm/extensions/008"  # Sponsor approval info location


class StudyVersion(ApiBaseModelWithId):
    versionIdentifier: str
    rationale: str
    documentVersionIds: List[str] = []
    dateValues: List[GovernanceDate] = []
    amendments: List[StudyAmendment] = []
    businessTherapeuticAreas: List[Code] = []
    studyIdentifiers: List[StudyIdentifier]
    referenceIdentifiers: List[ReferenceIdentifier] = []
    studyDesigns: List[Union[InterventionalStudyDesign, ObservationalStudyDesign]] = []
    titles: List[StudyTitle]
    eligibilityCriterionItems: List[EligibilityCriterionItem] = []
    narrativeContentItems: List[NarrativeContentItem] = []
    abbreviations: List[Abbreviation] = []
    roles: List[StudyRole] = []
    organizations: List[Organization] = []
    studyInterventions: List[StudyIntervention] = []
    administrableProducts: List[AdministrableProduct] = []
    medicalDevices: List[MedicalDevice] = []
    productOrganizationRoles: List[ProductOrganizationRole] = []
    biomedicalConcepts: List[BiomedicalConcept] = []
    bcCategories: List[BiomedicalConceptCategory] = []
    bcSurrogates: List[BiomedicalConceptSurrogate] = []
    dictionaries: List[SyntaxTemplateDictionary] = []
    conditions: List[Condition] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyVersion"]

    def confidentiality_statement(self) -> str:
        ext: Extension = self.get_extension(CS_EXT_URL)
        return ext.valueString if ext else ""

    def original_version(self) -> bool:
        ext: Extension = self.get_extension(OV_EXT_URL)
        return ext.valueBoolean if ext else False

    def first_amendment(self) -> StudyAmendment | None:
        if self.amendments:
            all_items = {x.id: x for x in self.amendments}
            all_ids = list(all_items.keys())
            prev_ids = [x.previousId for x in self.amendments]
            first = list(set(all_ids) - set(prev_ids))
            if first:
                return all_items[first[0]]
        return None

    def get_title(self, title_type):
        for title in self.titles:
            if title.type.decode == title_type:
                return title
        return None

    def sponsor_organization(self) -> Organization | None:
        org = self._find_first_organization("C70793")
        if org is None:
            # Fallback, find a sponsor organization, temporary
            org = next((x for x in self.organizations if x.type.code == "C54149"), None)
        return org

    def sponsor_identifier(self) -> StudyIdentifier | None:
        org = self.sponsor_organization()
        if org:
            for identifier in self.studyIdentifiers:
                if identifier.scopeId == org.id:
                    return identifier
        return None

    @deprecated("Use sponsor_organization method")
    def sponsor(self) -> Organization:
        map = self.organization_map()
        for x in self.studyIdentifiers:
            if x.is_sponsor(map):
                return map[x.scopeId]
        return None

    def sponsor_identifier_text(self) -> StudyIdentifier:
        identifier = self.sponsor_identifier()
        return identifier.text if identifier else ""

    def sponsor_name(self) -> str:
        org: Organization = self.sponsor_organization()
        return org.name if org else ""

    def sponsor_label(self) -> str:
        org: Organization = self.sponsor_organization()
        return org.label if org else ""

    def sponsor_label_name(self) -> str:
        label = self.sponsor_label()
        return label if label else self.sponsor_name()

    def sponsor_address(self) -> str:
        org: Organization = self.sponsor_organization()
        if org:
            return org.legalAddress.text if org.legalAddress else ""
        return ""

    def co_sponsor_organization(self) -> Organization | None:
        return self._find_first_organization("C215669")

    def co_sponsor_name(self) -> str:
        org: Organization = self.co_sponsor_organization()
        return org.name if org else ""

    def co_sponsor_label(self) -> str:
        org: Organization = self.co_sponsor_organization()
        return org.label if org else ""

    def co_sponsor_label_name(self) -> str:
        label = self.co_sponsor_label()
        return label if label else self.co_sponsor_name()

    def co_sponsor_address(self) -> str:
        org: Organization = self.co_sponsor_organization()
        if org:
            return org.legalAddress.text if org.legalAddress else ""
        return ""

    def local_sponsor_organization(self) -> Organization | None:
        return self._find_first_organization("C215670")

    def local_sponsor_name(self) -> str:
        org: Organization = self.local_sponsor_organization()
        return org.name if org else ""

    def local_sponsor_label(self) -> str:
        org: Organization = self.local_sponsor_organization()
        return org.label if org else ""

    def local_sponsor_label_name(self) -> str:
        label = self.local_sponsor_label()
        return label if label else self.local_sponsor_name()

    def local_sponsor_address(self) -> str:
        org: Organization = self.local_sponsor_organization()
        if org:
            return org.legalAddress.text if org.legalAddress else ""
        return ""

    def manufacturer_organization(self) -> Organization | None:
        return self._find_first_organization("C25392")

    def device_manufacturer_organization(self) -> Organization | None:
        manufacturers: list[Organization] = self._find_organizations("C25392")
        return next((x for x in manufacturers if x.type.code == "C215661"), None)

    def device_manufacturer_name(self) -> str:
        org: Organization = self.device_manufacturer_organization()
        return org.name if org else ""

    def device_manufacturer_label(self) -> str:
        org: Organization = self.device_manufacturer_organization()
        return org.label if org else ""

    def device_manufacturer_label_name(self) -> str:
        label = self.device_manufacturer_label()
        return label if label else self.device_manufacturer_name()

    def device_manufacturer_address(self) -> str:
        org: Organization = self.device_manufacturer_organization()
        if org:
            return org.legalAddress.text if org.legalAddress else ""
        return ""

    def _find_first_organization(self, role_code: str) -> Organization | None:
        orgs = self._find_organizations(role_code)
        return orgs[0] if len(orgs) > 0 else None

    def _find_organizations(self, role_code: str) -> list[Organization]:
        role = next((x for x in self.roles if x.code.code == role_code), None)
        return [self.organization(x) for x in role.organizationIds] if role else []

    def regulatory_identifiers(self) -> list[StudyIdentifier]:
        results = []
        for identifier in self.studyIdentifiers:
            org = self.organization(identifier.scopeId)
            if org and org.type.code == "C188863":
                results.append(identifier)
        return results

    def registry_identifiers(self) -> list[StudyIdentifier]:
        results = []
        for identifier in self.studyIdentifiers:
            org = self.organization(identifier.scopeId)
            if org and org.type.code == "C93453":
                results.append(identifier)
        return results

    def organization(self, id: str) -> Organization:
        return next((x for x in self.organizations if x.id == id), None)

    def criterion_item(self, id: str) -> EligibilityCriterionItem:
        return next((x for x in self.eligibilityCriterionItems if x.id == id), None)

    def intervention(self, id: str) -> StudyIntervention:
        return next((x for x in self.studyInterventions if x.id == id), None)

    def condition(self, id: str) -> Condition:
        return next((x for x in self.conditions if x.id == id), None)

    def phases(self) -> str:
        return ", ".join([sd.phase_as_text() for sd in self.studyDesigns])

    def official_title_text(self) -> str:
        for x in self.titles:
            if x.is_official():
                return x.text
        return ""

    def short_title_text(self) -> str:
        for x in self.titles:
            if x.is_short():
                return x.text
        return ""

    def acronym_text(self) -> str:
        for x in self.titles:
            if x.is_acronym():
                return x.text
        return ""

    def official_title(self) -> StudyIdentifier:
        for x in self.titles:
            if x.is_official():
                return x
        return None

    def short_title(self) -> StudyIdentifier:
        for x in self.titles:
            if x.is_short():
                return x
        return None

    def acronym(self) -> StudyIdentifier:
        for x in self.titles:
            if x.is_acronym():
                return x
        return None

    def nct_identifier(self) -> StudyIdentifier:
        map = self.organization_map()
        for x in self.studyIdentifiers:
            if map[x.scopeId].name == "CT.GOV":
                return x.text
        return ""

    def fda_ind_identifier(self) -> StudyIdentifier:
        map = self.organization_map()
        for x in self.studyIdentifiers:
            if map[x.scopeId].name == "FDA":
                return x.text
        return ""

    def ema_identifier(self) -> StudyIdentifier:
        map = self.organization_map()
        for x in self.studyIdentifiers:
            if map[x.scopeId].name == "EMA":
                return x.text
        return ""

    @deprecated("Use the method relating to the document")
    def protocol_date(self) -> GovernanceDate:
        for x in self.dateValues:
            if x.type.code == "C71476":
                return x
        return ""

    def approval_date(self) -> GovernanceDate | None:
        for x in self.dateValues:
            if x.type.code == "C71476":
                return x
        return None

    @deprecated("Use the method relating to the document")
    def protocol_date_value(self) -> date:
        for x in self.dateValues:
            if x.type.code == "C71476":
                return x.dateValue
        return ""

    def approval_date_value(self) -> date | None:
        for x in self.dateValues:
            if x.type.code == "C71476":
                return x.dateValue
        return None

    def approval_date_text(self) -> str | None:
        for x in self.dateValues:
            if x.type.code == "C71476":
                return x.dateValue.strftime("%Y-%m-%d")
        return None

    def sponsor_approval_location(self) -> str:
        ext: Extension = self.get_extension(SAL_EXT_URL)
        return ext.valueString if ext else ""

    def find_study_design(self, id: str) -> StudyDesign:
        return next((x for x in self.studyDesigns if x.id == id), None)

    def documents(
        self,
        document_map: dict[
            str, dict[StudyDefinitionDocument, StudyDefinitionDocumentVersion]
        ],
    ) -> list[dict[StudyDefinitionDocument, StudyDefinitionDocumentVersion]]:
        return [document_map[x] for x in self.documentVersionIds]

    def to_html(
        self,
        template: str,
        document_map: dict[
            str, dict[StudyDefinitionDocument, StudyDefinitionDocumentVersion]
        ],
    ) -> str | None:
        sddv: StudyDefinitionDocumentVersion
        if sddv := self.study_document_version(template, document_map):
            return sddv.to_html(self.narrative_content_item_map())
        return None

    def study_document_version(
        self,
        template: str,
        document_map: dict[
            str, dict[StudyDefinitionDocument, StudyDefinitionDocumentVersion]
        ],
    ) -> StudyDefinitionDocumentVersion | None:
        docs_info = self.documents(document_map)
        for doc_info in docs_info:
            study_document: StudyDefinitionDocument = doc_info["document"]
            if study_document.templateName.upper() == template.upper():
                return doc_info["version"]
        return None

    def compound_codes(self) -> str:
        ext: Extension = self.get_extension(CC_EXT_URL)
        return ext.valueString if ext else ""

    def compound_names(self) -> str:
        ext: Extension = self.get_extension(CN_EXT_URL)
        return ext.valueString if ext else ""

    def medical_expert(self) -> str:
        ext: Extension = self.get_extension(ME_EXT_URL)
        return ext.valueString if ext else ""

    def sponsor_signatory(self) -> str:
        ext: Extension = self.get_extension(SS_EXT_URL)
        return ext.valueString if ext else ""

    def role_map(self) -> dict[str, StudyRole]:
        return {x.id: x for x in self.roles}

    def organization_map(self) -> dict[str, Organization]:
        return {x.id: x for x in self.organizations}

    def eligibility_critieria_item_map(self) -> dict[str, EligibilityCriterionItem]:
        return {x.id: x for x in self.eligibilityCriterionItems}

    def narrative_content_item_map(self) -> dict[str, NarrativeContentItem]:
        return {x.id: x for x in self.narrativeContentItems}



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00024(RuleTemplate):
    """
    DDF00024: An epoch must only reference epochs that are specified within the same study design.

    Applies to: StudyEpoch
    Attributes: previous, next
    """

    def __init__(self):
        super().__init__(
            "DDF00024",
            RuleTemplate.ERROR,
            "An epoch must only reference epochs that are specified within the same study design.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .document_content_reference import DocumentContentReference


class StudyChange(ApiBaseModelWithIdNameLabelAndDesc):
    summary: str
    rationale: str
    changedSections: List[DocumentContentReference]
    instanceType: Literal["StudyChange"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00065(RuleTemplate):
    """
    DDF00065: A scheduled decision instance is not expected to have a sub-timeline.

    Applies to: ScheduledDecisionInstance
    Attributes: timeline
    """

    def __init__(self):
        super().__init__(
            "DDF00065",
            RuleTemplate.ERROR,
            "A scheduled decision instance is not expected to have a sub-timeline.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .code import Code
from .address import Address
from .study_site import StudySite


class Organization(ApiBaseModelWithIdNameAndLabel):
    type: Code
    identifierScheme: str
    identifier: str
    legalAddress: Union[Address, None] = None
    managedSites: List[StudySite] = []
    instanceType: Literal["Organization"]



from typing import Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc


class BiospecimenRetention(ApiBaseModelWithIdNameLabelAndDesc):
    isRetained: bool
    includesDNA: Union[bool, None] = None
    instanceType: Literal["BiospecimenRetention"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00081(RuleTemplate):
    """
    DDF00081: Class relationships must conform with the USDM schema based on the API specification.

    Applies to: All
    Attributes: All
    """

    def __init__(self):
        super().__init__(
            "DDF00081",
            RuleTemplate.ERROR,
            "Class relationships must conform with the USDM schema based on the API specification.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .quantity_range import Quantity
from .geographic_scope import GeographicScope


class SubjectEnrollment(ApiBaseModelWithIdNameLabelAndDesc):
    quantity: Quantity
    forGeographicScope: Union[GeographicScope, None] = None
    forStudyCohortId: Union[str, None] = None
    forStudySiteId: Union[str, None] = None
    instanceType: Literal["SubjectEnrollment"]



from typing import Literal
from .api_base_model import ApiBaseModelWithId


class Code(ApiBaseModelWithId):
    code: str
    codeSystem: str
    codeSystemVersion: str
    decode: str
    instanceType: Literal["Code"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0240(RuleTemplate):
    """
    CHK0240: There must be a one-to-one relationship between referenced section number and title within a study amendment.

    Applies to: DocumentContentReference
    Attributes: sectionNumber, sectionTitle
    """

    def __init__(self):
        super().__init__(
            "CHK0240",
            RuleTemplate.ERROR,
            "There must be a one-to-one relationship between referenced section number and title within a study amendment.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00040(RuleTemplate):
    """
    DDF00040: Each study element must be referenced by at least one study cell.

    Applies to: StudyCell
    Attributes: elements
    """

    def __init__(self):
        super().__init__(
            "DDF00040",
            RuleTemplate.ERROR,
            "Each study element must be referenced by at least one study cell.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .code import Code


class StudyAmendmentReason(ApiBaseModelWithId):
    code: Code
    otherReason: Union[str, None] = None
    instanceType: Literal["StudyAmendmentReason"]

    def reason_as_text(self) -> str:
        return self.code.decode

    def other_reason_as_text(self) -> str:
        return self.otherReason if self.otherReason else ""



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0197(RuleTemplate):
    """
    CHK0197: An identified organization is not expected to have more than 1 identifier for the study.

    Applies to: StudyIdentifier
    Attributes: scope
    """

    def __init__(self):
        super().__init__(
            "CHK0197",
            RuleTemplate.ERROR,
            "An identified organization is not expected to have more than 1 identifier for the study.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0232(RuleTemplate):
    """
    CHK0232: A planned sex must ether include a single entry of male or female or both female and male as entries.

    Applies to: StudyDesignPopulation, StudyCohort
    Attributes: plannedSex
    """

    def __init__(self):
        super().__init__(
            "CHK0232",
            RuleTemplate.ERROR,
            "A planned sex must ether include a single entry of male or female or both female and male as entries.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.rules_validation import RulesValidationEngine
from usdm3.base.singleton import Singleton


class RulesValidation4(metaclass=Singleton):
    def __init__(self, root_path: str):
        self.rules_validation = RulesValidationEngine(root_path, "usdm4.rules.library")

    def validate(self, filename: str):
        return self.rules_validation.validate_rules(filename)



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00048(RuleTemplate):
    """
    DDF00048: A procedure must only reference a study intervention that is defined within the same study design as the activity within which the procedure is defined.

    Applies to: Procedure
    Attributes: studyIntervention
    """

    def __init__(self):
        super().__init__(
            "DDF00048",
            RuleTemplate.ERROR,
            "A procedure must only reference a study intervention that is defined within the same study design as the activity within which the procedure is defined.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import re
import operator
from usdm4.api.study_design import StudyDesign
from usdm4.api.schedule_timeline import ScheduleTimeline
from usdm4.api.scheduled_instance import (
    ScheduledActivityInstance,
    ScheduledDecisionInstance,
    ScheduledInstance,
    ConditionAssignment,
)
from usdm4.api.schedule_timeline_exit import ScheduleTimelineExit
from simple_error_log import Errors
from simple_error_log.error_location import KlassMethodLocation
from .timepoint import Timepoint


class Expander:
    MODULE = "usdm4.expander.expander.Expander"

    def __init__(
        self, study_design: StudyDesign, timeline: ScheduleTimeline, errors: Errors
    ):
        # self._path = None
        self._study_design = study_design
        self._timeline = timeline
        self._errors = errors
        self._id = 1
        self._nodes: list[Timepoint] = []

    @property
    def nodes(self):
        return self._nodes

    def process(self):
        entry: ScheduledInstance = self._timeline.find_timepoint(self._timeline.entryId)
        self._process_si(self._timeline, entry, 0)
        node: Timepoint
        new_nodes = []
        for node in self._nodes:
            # print(f"NODE CLEAN: {node.id}")
            if node.activities:
                # print("NODE KEEP")
                new_nodes.append(node)
        self._nodes = sorted(new_nodes, key=lambda x: x.tick)
        # print(f"EXPANDED VIEW: {[x.to_dict() for x in self._nodes]}")

    def _process_si(
        self,
        timeline: ScheduleTimeline,
        si: ScheduledActivityInstance
        | ScheduledDecisionInstance
        | ScheduleTimelineExit,
        offset: int,
    ):
        # print(f"SI {si.id}, {si.instanceType}, {type(si)}, {offset}")
        if isinstance(si, ScheduledActivityInstance):
            # print(f"SAI with id {si.id}")
            tp = Timepoint(
                self._study_design,
                timeline,
                si,
                self._errors,
                self._id,
                offset if not timeline.mainTimeline else 0,
            )
            self._id += 1
            self._nodes.append(tp)

            # Timepoint timeline
            if si.timelineId:
                # print(f"Timepoint timeline {si.timelineId}")
                tp_timeline = self._study_design.find_timeline(si.timelineId)
                entry: ScheduledInstance = tp_timeline.find_timepoint(
                    tp_timeline.entryId
                )
                self._process_si(tp_timeline, entry, tp.tick)

            # Activity timelines
            a_timelines = tp.activity_timelines()
            for a_timeline in a_timelines:
                # print(f"ACTIVITY TIMELINE: {a_timeline.id}")
                entry: ScheduledInstance = a_timeline.find_timepoint(a_timeline.entryId)
                self._process_si(a_timeline, entry, tp.tick)

            # Next
            if si.defaultConditionId:
                self._process_si(
                    timeline, timeline.find_timepoint(si.defaultConditionId), tp.tick
                )
            elif si.timelineExitId:
                pass
                # self._process_si(
                #     timeline, timeline.find_exit(si.timelineExitId), tp.tick
                # )
            else:
                self._errors.error(
                    f"Next instance error, {si}",
                    KlassMethodLocation(self.MODULE, "_process_si"),
                )
        elif isinstance(si, ScheduledDecisionInstance):
            # print(f"SDI with id {si.id}")
            if len(si.conditionAssignments) == 1:
                ca: ConditionAssignment = si.conditionAssignments[0]
                dc_op, dc_value = self._days_condition(ca.condition)
                if dc_op:
                    # print(f"DECISION: {offset}, {dc_value * 24 * 60 * 60}, {dc_op}")
                    if dc_op(offset, dc_value * 24 * 60 * 60):
                        self._process_si(
                            timeline,
                            timeline.find_timepoint(ca.conditionTargetId),
                            offset,
                        )
                    else:
                        self._process_si(
                            timeline,
                            timeline.find_timepoint(si.defaultConditionId),
                            offset,
                        )
                else:
                    self._errors.error(
                        "No day condition encountered, being ignored.",
                        KlassMethodLocation(self.MODULE, "_process_si"),
                    )
                    self._process_si(
                        timeline,
                        timeline.find_timepoint(si.defaultConditionId),
                        offset,
                    )
            else:
                self._errors.error(
                    "Complex condition encountered, being ignored.",
                    KlassMethodLocation(self.MODULE, "_process_si"),
                )
                self._process_si(
                    timeline,
                    timeline.find_timepoint(si.defaultConditionId),
                    offset,
                )
        elif isinstance(si, ScheduleTimelineExit):
            pass
        else:
            self._errors.error(
                f"Unknown instance type detected, {si}",
                KlassMethodLocation(self.MODULE, "_process_si"),
            )

    def _days_condition(self, text) -> tuple[object, int]:
        try:
            operators = {">": operator.gt, "<": operator.lt, "=": operator.eq}
            pattern = r"(?i)days?\s*([<>=])\s*(\d+)"
            match = re.search(pattern, text)
            if match:
                op = operators[match.group(1)]  # Select the operator
                value = int(match.group(2))
                return op, value
            return None, None
        except Exception as e:
            self._errors.exception(
                f"Error detected processing, '{text}'",
                e,
                KlassMethodLocation(self.MODULE, "_days_condition"),
            )
            return None, None



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0208(RuleTemplate):
    """
    CHK0208: Within a study protocol document version, if a date of a specific type exists with a global geographic scope then no other dates are expected with the same type.

    Applies to: StudyDefinitionDocumentVersion
    Attributes: dateValues
    """

    def __init__(self):
        super().__init__(
            "CHK0208",
            RuleTemplate.ERROR,
            "Within a study protocol document version, if a date of a specific type exists with a global geographic scope then no other dates are expected with the same type.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .code import Code
from .comment_annotation import CommentAnnotation


class StudyEpoch(ApiBaseModelWithIdNameLabelAndDesc):
    type: Code
    previousId: Union[str, None] = None
    nextId: Union[str, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyEpoch"]



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .study_amendment_reason import StudyAmendmentReason
from .study_change import StudyChange
from .study_amendment_impact import StudyAmendmentImpact
from .geographic_scope import GeographicScope
from .subject_enrollment import SubjectEnrollment
from .governance_date import GovernanceDate
from .comment_annotation import CommentAnnotation
from .extension import ExtensionAttribute

SI_EXT_URL = (
    "www.d4k.dk/usdm/extensions/003"  # Site identifier scope. An array of extensions.
)


class StudyAmendment(ApiBaseModelWithIdNameLabelAndDesc):
    number: str
    summary: str
    primaryReason: StudyAmendmentReason
    secondaryReasons: List[StudyAmendmentReason] = []
    changes: List[StudyChange] = []  # Not in API
    impacts: List[StudyAmendmentImpact] = []
    geographicScopes: List[GeographicScope]
    enrollments: List[SubjectEnrollment] = []
    dateValues: List[GovernanceDate] = []
    previousId: Union[str, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyAmendment"]

    def site_identifier_scopes(self) -> list[str]:
        ext: ExtensionAttribute = self.get_extension(SI_EXT_URL)
        return [x.valueString for x in ext.extensionAttributes] if ext else []

    def site_identifier_scopes_as_text(self) -> str:
        return (",").join(self.site_identifier_scopes())

    def primary_reason_as_text(self) -> str:
        return self.primaryReason.reason_as_text()

    def primary_other_reason_as_text(self) -> str:
        return self.primaryReason.other_reason_as_text()

    def secondary_reason_as_text(self) -> str:
        return (
            self.secondaryReasons[0].reason_as_text()
            if len(self.secondaryReasons) > 0
            else ""
        )

    def secondary_other_reason_as_text(self) -> str:
        return (
            self.secondaryReasons[0].other_reason_as_text()
            if len(self.secondaryReasons) > 0
            else ""
        )

    def is_global(self) -> bool:
        for scope in self.geographicScopes:
            if scope.type.code == "C68846":
                return True
        return False



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0238(RuleTemplate):
    """
    CHK0238: At least one attribute must be specified for an address.

    Applies to: Address
    Attributes: All
    """

    def __init__(self):
        super().__init__(
            "CHK0238",
            RuleTemplate.ERROR,
            "At least one attribute must be specified for an address.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0257(RuleTemplate):
    """
    CHK0257: If the intervention model indicates a single group design then only one intervention is expected. In all other cases more interventions are expected.

    Applies to: InterventionalStudyDesign
    Attributes: model
    """

    def __init__(self):
        super().__init__(
            "CHK0257",
            RuleTemplate.ERROR,
            "If the intervention model indicates a single group design then only one intervention is expected. In all other cases more interventions are expected.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0271(RuleTemplate):
    """
    CHK0271: An interventional study must be specified using the InterventionalStudyDesign class.

    Applies to: InterventionalStudyDesign
    Attributes: studyType
    """

    def __init__(self):
        super().__init__(
            "CHK0271",
            RuleTemplate.ERROR,
            "An interventional study must be specified using the InterventionalStudyDesign class. ",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0170(RuleTemplate):
    """
    CHK0170: Within an encounter, if more environmental settings are defined, they must be distinct.

    Applies to: Encounter
    Attributes: environmentalSettings
    """

    def __init__(self):
        super().__init__(
            "CHK0170",
            RuleTemplate.ERROR,
            "Within an encounter, if more environmental settings are defined, they must be distinct.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0234(RuleTemplate):
    """
    CHK0234: A study role must not reference both assigned persons and organizations.

    Applies to: StudyRole
    Attributes: assignedPersons, organizations
    """

    def __init__(self):
        super().__init__(
            "CHK0234",
            RuleTemplate.ERROR,
            "A study role must not reference both assigned persons and organizations.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc, ApiBaseModelWithId


class ParameterMap(ApiBaseModelWithId):
    tag: str
    reference: str
    instanceType: Literal["ParameterMap"]


class SyntaxTemplateDictionary(ApiBaseModelWithIdNameLabelAndDesc):
    parameterMaps: List[ParameterMap] = []  # Allow for empty list, not in API
    instanceType: Literal["SyntaxTemplateDictionary"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0245(RuleTemplate):
    """
    CHK0245: There must be exactly one study role with a code of sponsor.

    Applies to: StudyRole
    Attributes: code
    """

    def __init__(self):
        super().__init__(
            "CHK0245",
            RuleTemplate.ERROR,
            "There must be exactly one study role with a code of sponsor. ",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0233(RuleTemplate):
    """
    CHK0233: Every study role must apply to either a study version or at least one study design, but not both.

    Applies to: StudyRole
    Attributes: appliesTo
    """

    def __init__(self):
        super().__init__(
            "CHK0233",
            RuleTemplate.ERROR,
            "Every study role must apply to either a study version or at least one study design, but not both.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code


class StudyTitle(ApiBaseModelWithId):
    text: str
    type: Code
    instanceType: Literal["StudyTitle"]

    def is_official(self) -> bool:
        return True if self.type.code == "C207616" else False

    def is_acronym(self) -> bool:
        return True if self.type.code == "C207646" else False

    def is_short(self) -> bool:
        return True if self.type.code == "C207615" else False



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00038(RuleTemplate):
    """
    DDF00038: When included in text, references to items specified in the dictionary must be specified in the correct format. They must start with <usdm:tag, end with either '/>'', and must contain name=\"parametername\" (where \"parametername\"  contain only letters, numbers or underscores).

    Applies to: EligibilityCriterion, Characteristic, Condition, Objective, Endpoint
    Attributes: text
    """

    def __init__(self):
        super().__init__(
            "DDF00038",
            RuleTemplate.ERROR,
            'When included in text, references to items specified in the dictionary must be specified in the correct format. They must start with <usdm:tag, end with either \'/>\'\', and must contain name="parametername" (where "parametername"  contain only letters, numbers or underscores).',
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal, Union, List
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .syntax_template import SyntaxTemplate
from .comment_annotation import CommentAnnotation
from .code import Code


class EligibilityCriterion(ApiBaseModelWithIdNameLabelAndDesc):
    category: Code
    identifier: str
    criterionItemId: str
    nextId: Union[str, None] = None
    previousId: Union[str, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["EligibilityCriterion"]


class EligibilityCriterionItem(SyntaxTemplate):
    instanceType: Literal["EligibilityCriterionItem"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00096(RuleTemplate):
    """
    DDF00096: All primary endpoints must be referenced by a primary objective.

    Applies to: Endpoint
    Attributes: level
    """

    def __init__(self):
        super().__init__(
            "DDF00096",
            RuleTemplate.ERROR,
            "All primary endpoints must be referenced by a primary objective.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00125(RuleTemplate):
    """
    DDF00125: Attributes must be included as defined in the USDM schema based on the API specification (i.e., all required properties are present and no additional attributes are present).

    Applies to: All
    Attributes: All
    """

    def __init__(self):
        super().__init__(
            "DDF00125",
            RuleTemplate.ERROR,
            "Attributes must be included as defined in the USDM schema based on the API specification (i.e., all required properties are present and no additional attributes are present).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .code import Code


class ResponseCode(ApiBaseModelWithIdNameAndLabel):
    isEnabled: bool
    code: Code
    instanceType: Literal["ResponseCode"]



import os
import pathlib
from usdm3.rules.library.rule_ddf00082 import RuleDDF00082 as V3Rule
from usdm3.rules.library.schema.schema_location import SchemaErrorLocation
from usdm3.rules.library.schema.schema_validation import (
    SchemaValidation,
    ValidationError,
)
from usdm3.data_store.data_store import DataStore


class RuleDDF00082(V3Rule):
    def validate(self, config: dict) -> bool:
        try:
            data: DataStore = config["data"]
            path = self._schema_path()
            validator = SchemaValidation(path)
            validator.validate_file(data.filename, "Wrapper-Input")
            return True
        except ValidationError as e:
            location = SchemaErrorLocation(e.json_path, e.instance)
            self._errors.add(f"Message: {e.message}\nContext: {e.context}", location)
            return False

    def _schema_path(self) -> str:
        root = pathlib.Path(__file__).parent.resolve()
        return os.path.join(root, "schema/usdm_v4-0-0.json")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdAndName


class NarrativeContentItem(ApiBaseModelWithIdAndName):
    text: str
    instanceType: Literal["NarrativeContentItem"]


class NarrativeContent(ApiBaseModelWithIdAndName):
    sectionNumber: Union[str, None] = None
    sectionTitle: Union[str, None] = None
    displaySectionNumber: bool
    displaySectionTitle: bool
    childIds: List[str] = []
    previousId: Union[str, None] = None
    nextId: Union[str, None] = None
    contentItemId: Union[str, None] = None
    instanceType: Literal["NarrativeContent"]

    def content(self, narrative_content_item_map: dict) -> str | None:
        content: NarrativeContentItem = self.content_item(narrative_content_item_map)
        return self.format_heading() + content.text if content else None

    def content_item(
        self, narrative_content_item_map: dict
    ) -> NarrativeContentItem | None:
        return (
            narrative_content_item_map[self.contentItemId]
            if self.contentItemId in narrative_content_item_map
            else None
        )

    def level(self) -> int:
        if self.sectionNumber is None:
            result = 1
        elif self.sectionNumber.lower().startswith("appendix"):
            result = 1
        else:
            text = (
                self.sectionNumber[:-1]
                if self.sectionNumber.endswith(".")
                else self.sectionNumber
            )
            result = len(text.split("."))
        return result

    def format_heading(self) -> str:
        level = self.level()
        number = self.sectionNumber
        title = self.sectionTitle
        if number and number == "0":
            return ""
        elif number and title:
            return f"<h{level}>{number} {title}</h{level}>"
        elif number:
            return f"<h{level}>{number}</h{level}>"
        elif title:
            return f"<h{level}>{title}</h{level}>"
        else:
            return ""



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .procedure import Procedure
from .comment_annotation import CommentAnnotation


class Activity(ApiBaseModelWithIdNameLabelAndDesc):
    previousId: Union[str, None] = None
    nextId: Union[str, None] = None
    childIds: List[str] = []
    definedProcedures: List[Procedure] = []
    biomedicalConceptIds: List[str] = []
    bcCategoryIds: List[str] = []
    bcSurrogateIds: List[str] = []
    timelineId: Union[str, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["Activity"]



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .quantity_range import Quantity
from .administration import Administration
from .code import Code
from .comment_annotation import CommentAnnotation


class StudyIntervention(ApiBaseModelWithIdNameLabelAndDesc):
    role: Code
    type: Code
    minimumResponseDuration: Union[Quantity, None] = None
    codes: List[Code] = []
    administrations: List[Administration] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyIntervention"]



from typing import List, Literal
from .syntax_template import SyntaxTemplate
from .code import Code
from .endpoint import Endpoint


class Objective(SyntaxTemplate):
    level: Code
    endpoints: List[Endpoint] = []
    instanceType: Literal["Objective"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0256(RuleTemplate):
    """
    CHK0256: If 'appliesTo' is specified for a product organization role, then the product organization role must only apply to medical devices or administrable products.

    Applies to: ProductOrganizationRole
    Attributes: appliesTo
    """

    def __init__(self):
        super().__init__(
            "CHK0256",
            RuleTemplate.ERROR,
            "If 'appliesTo' is specified for a product organization role, then the product organization role must only apply to medical devices or administrable products.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00152(RuleTemplate):
    """
    DDF00152: An activity must only reference timelines that are specified within the same study design.

    Applies to: Activity
    Attributes: timeline
    """

    def __init__(self):
        super().__init__(
            "DDF00152",
            RuleTemplate.ERROR,
            "An activity must only reference timelines that are specified within the same study design.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0219(RuleTemplate):
    """
    CHK0219: A referenced substance must not have any references itself.

    Applies to: Substance
    Attributes: referenceSubstance
    """

    def __init__(self):
        super().__init__(
            "CHK0219",
            RuleTemplate.ERROR,
            "A referenced substance must not have any references itself.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc, ApiBaseModelWithId


class ConditionAssignment(ApiBaseModelWithId):
    condition: str
    conditionTargetId: str
    instanceType: Literal["ConditionAssignment"]


class ScheduledInstance(ApiBaseModelWithIdNameLabelAndDesc):
    defaultConditionId: Union[str, None] = None
    epochId: Union[str, None] = None
    instanceType: Literal["ScheduledInstance"]


class ScheduledActivityInstance(ScheduledInstance):
    timelineId: Union[str, None] = None
    timelineExitId: Union[str, None] = None
    activityIds: List[str] = []
    encounterId: Union[str, None] = None
    instanceType: Literal["ScheduledActivityInstance"]


class ScheduledDecisionInstance(ScheduledInstance):
    conditionAssignments: List[
        ConditionAssignment
    ]  # = []  # Allow for empty list, not in API
    instanceType: Literal["ScheduledDecisionInstance"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0179(RuleTemplate):
    """
    CHK0179: The ordering of activities (using the previous and next attributes) must include the parents (e.g. activities refering to children) preceding their children.

    Applies to: Activity
    Attributes: previous, next
    """

    def __init__(self):
        super().__init__(
            "CHK0179",
            RuleTemplate.ERROR,
            "The ordering of activities (using the previous and next attributes) must include the parents (e.g. activities refering to children) preceding their children. ",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00132(RuleTemplate):
    """
    DDF00132: Within a study design, if a planned completion number is defined, it must be specified either in the study population or in all cohorts.

    Applies to: StudyDesignPopulation, StudyCohort
    Attributes: plannedCompletionNumber
    """

    def __init__(self):
        super().__init__(
            "DDF00132",
            RuleTemplate.ERROR,
            "Within a study design, if a planned completion number is defined, it must be specified either in the study population or in all cohorts.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00064(RuleTemplate):
    """
    DDF00064: A scheduled decision instance is not expected to refer to a timeline exit.

    Applies to: ScheduledDecisionInstance
    Attributes: timelineExit
    """

    def __init__(self):
        super().__init__(
            "DDF00064",
            RuleTemplate.ERROR,
            "A scheduled decision instance is not expected to refer to a timeline exit.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0004(RuleTemplate):
    """
    CHK0004: Within a study design, there must be at least 1 eligibility criterion that is referenced by either a study population or a cohort.

    Applies to: StudyDesignPopulation, StudyCohort
    Attributes: criteria
    """

    def __init__(self):
        super().__init__(
            "CHK0004",
            RuleTemplate.ERROR,
            "Within a study design, there must be at least 1 eligibility criterion that is referenced by either a study population or a cohort.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00074(RuleTemplate):
    """
    DDF00074: If the intervention model indicates a single group design then only one intervention is expected. In all other cases more interventions are expected.

    Applies to: StudyDesign
    Attributes: studyInterventions
    """

    def __init__(self):
        super().__init__(
            "DDF00074",
            RuleTemplate.ERROR,
            "If the intervention model indicates a single group design then only one intervention is expected. In all other cases more interventions are expected.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0249(RuleTemplate):
    """
    CHK0249: An administrable product must not be referenced as both the administrable product for an administration and the embedded product of a medical device that is referenced by the same administration.

    Applies to: Administration
    Attributes: administrableProduct
    """

    def __init__(self):
        super().__init__(
            "CHK0249",
            RuleTemplate.ERROR,
            "An administrable product must not be referenced as both the administrable product for an administration and the embedded product of a medical device that is referenced by the same administration.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal
from typing_extensions import deprecated
from .api_base_model import ApiBaseModelWithId
from .organization import Organization
from .code import Code


class Identifier(ApiBaseModelWithId):
    text: str
    scopeId: str
    instanceType: Literal["Identifier"]


class ReferenceIdentifier(Identifier):
    type: Code
    instanceType: Literal["ReferenceIdentifier"]


class StudyIdentifier(Identifier):
    instanceType: Literal["StudyIdentifier"]

    @deprecated("Use roles not identifiers to find a sponsor organization")
    def is_sponsor(self, organization_map: dict) -> bool:
        org = organization_map[self.scopeId]
        return True if org.type.code == "C54149" else False

    def scoped_by(self, organization_map: dict) -> Organization:
        return organization_map[self.scopeId]


class AdministrableProductIdentifier(Identifier):
    instanceType: Literal["AdministrableProductIdentifier"]


class MedicalDeviceIdentifier(Identifier):
    type: Code
    instanceType: Literal["MedicalDeviceIdentifier"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00012(RuleTemplate):
    """
    DDF00012: Within a study design, there must be exactly one scheduled timeline which identifies as the main Timeline.

    Applies to: ScheduleTimeline
    Attributes: mainTimeline
    """

    def __init__(self):
        super().__init__(
            "DDF00012",
            RuleTemplate.ERROR,
            "Within a study design, there must be exactly one scheduled timeline which identifies as the main Timeline.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00087(RuleTemplate):
    """
    DDF00087: Encounter ordering using previous and next attributes is expected to be consistent with the order of corresponding scheduled activity instances according to their specified default conditions.

    Applies to: Encounter
    Attributes: previous, next
    """

    def __init__(self):
        super().__init__(
            "DDF00087",
            RuleTemplate.ERROR,
            "Encounter ordering using previous and next attributes is expected to be consistent with the order of corresponding scheduled activity instances according to their specified default conditions.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0201(RuleTemplate):
    """
    CHK0201: If an administration's dose is specified then a corresponding route is expected and vice versa.

    Applies to: Administration
    Attributes: dose, route
    """

    def __init__(self):
        super().__init__(
            "CHK0201",
            RuleTemplate.ERROR,
            "If an administration's dose is specified then a corresponding route is expected and vice versa.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .study_definition_document_version import StudyDefinitionDocumentVersion
from .code import Code
from .comment_annotation import CommentAnnotation


class StudyDefinitionDocument(ApiBaseModelWithIdNameLabelAndDesc):
    language: Code
    type: Code
    templateName: str
    versions: List[StudyDefinitionDocumentVersion] = []
    childIds: List[str] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyDefinitionDocument"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0200(RuleTemplate):
    """
    CHK0200: An administration's route must be specfied according to the extensible Route of Administration Response (C66729) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: Administration
    Attributes: route
    """

    def __init__(self):
        super().__init__(
            "CHK0200",
            RuleTemplate.ERROR,
            "An administration's route must be specfied according to the extensible Route of Administration Response (C66729) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0236(RuleTemplate):
    """
    CHK0236: A masking is expected to be defined for at least two study roles in a study design with a double blind blinding schema.

    Applies to: StudyRole
    Attributes: masking
    """

    def __init__(self):
        super().__init__(
            "CHK0236",
            RuleTemplate.ERROR,
            "A masking is expected to be defined for at least two study roles in a study design with a double blind blinding schema.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00056(RuleTemplate):
    """
    DDF00056: Within a study design, if more therapeutic areas are defined, they must be distinct.

    Applies to: StudyDesign
    Attributes: therapeuticAreas
    """

    def __init__(self):
        super().__init__(
            "DDF00056",
            RuleTemplate.ERROR,
            "Within a study design, if more therapeutic areas are defined, they must be distinct.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00006(RuleTemplate):
    """
    DDF00006: Timing windows must be fully defined, if one of the window attributes (i.e., window label, window lower, and window upper) is defined then all must be specified.

    Applies to: Timing
    Attributes: windowLabel, windowLower, windowUpper
    """

    def __init__(self):
        super().__init__(
            "DDF00006",
            RuleTemplate.ERROR,
            "Timing windows must be fully defined, if one of the window attributes (i.e., window label, window lower, and window upper) is defined then all must be specified.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from datetime import date
from .api_base_model import ApiBaseModelWithId
from .code import Code
from .governance_date import GovernanceDate
from .narrative_content import NarrativeContent, NarrativeContentItem
from .comment_annotation import CommentAnnotation


class StudyDefinitionDocumentVersion(ApiBaseModelWithId):
    version: str
    status: Code
    dateValues: List[GovernanceDate] = []
    contents: List[NarrativeContent] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyDefinitionDocumentVersion"]

    def approval_date(self) -> GovernanceDate | None:
        for x in self.dateValues:
            if x.type.code == "C71476":
                return x
        return None

    def approval_date_value(self) -> date | None:
        for x in self.dateValues:
            if x.type.code == "C71476":
                return x.dateValue
        return None

    def approval_date_text(self) -> str | None:
        for x in self.dateValues:
            if x.type.code == "C71476":
                return x.dateValue.strftime("%Y-%m-%d")
        return None

    def narrative_content_in_order(self):
        sections = []
        narrative_content = self._first_narrative_content()
        while narrative_content:
            sections.append(narrative_content)
            narrative_content = next(
                (x for x in self.contents if x.id == narrative_content.nextId), None
            )
        return sections

    def find_narrative_content(self, id: str) -> NarrativeContent | None:
        map = self.narrative_content_map()
        return map[id] if id in map else None

    def find_narrative_content_by_number(
        self, section_number: str
    ) -> NarrativeContent | None:
        return next(
            (x for x in self.contents if x.sectionNumber == section_number), None
        )

    def find_narrative_content_by_title(
        self, section_title: str
    ) -> NarrativeContent | None:
        return next(
            (
                x
                for x in self.contents
                if x.sectionTitle.upper() == section_title.upper()
            ),
            None,
        )

    def to_html(
        self, narrative_content_item_map: dict[str, NarrativeContentItem]
    ) -> str:
        result = ""
        narrative_contents = self.narrative_content_in_order()
        narrative_content: NarrativeContent
        for narrative_content in narrative_contents:
            content = narrative_content.content(narrative_content_item_map)
            if content:
                result += content
        return result

    def narrative_content_map(self) -> dict[NarrativeContent]:
        return {x.id: x for x in self.contents}

    def _first_narrative_content(self) -> NarrativeContent:
        return next((x for x in self.contents if not x.previousId and x.nextId), None)



from usdm4.__info__ import __model_version__, __package_version__
from usdm4.api.wrapper import Wrapper


class Convert:
    @classmethod
    def convert(cls, data: dict) -> dict:
        wrapper = data
        study = wrapper["study"]

        # Chnage the wrapper details
        wrapper["usdmVersion"] = __model_version__
        wrapper["systemName"] = "Python USDM4 Package"
        wrapper["systemVersion"] = __package_version__

        # Change type of documents
        if study["documentedBy"]:
            study["documentedBy"]["instanceType"] = "StudyDefinitionDocument"
            study["documentedBy"]["language"] = {
                "id": "LangaugeCode_1",
                "code": "en",
                "codeSystem": "ISO 639-1",
                "codeSystemVersion": "2007",
                "decode": "English",
                "instanceType": "Code",
            }
            study["documentedBy"]["documentType"] = {
                "id": "DocumentTypeCode_1",
                "code": "C70817",
                "codeSystem": "cdisc.org",
                "codeSystemVersion": "2023-12-15",
                "decode": "Study Protocol",
                "instanceType": "Code",
            }
            study["documentedBy"]["type"] = {
                "id": "DocumentTypeCode_1",
                "code": "C70817",
                "codeSystem": "cdisc.org",
                "codeSystemVersion": "2023-12-15",
                "decode": "Protocol",
                "instanceType": "Code",
            }
            study["documentedBy"]["templateName"] = "Sponsor Study Protocol Template"
            for versions in study["documentedBy"]["versions"]:
                Convert._move(versions, "protocolStatus", "status")
                Convert._move(versions, "protocolVersion", "version")
            study["documentedBy"] = [study["documentedBy"]]
        else:
            study["documentedBy"] = []

        for version in study["versions"]:
            doc_id = version["documentVersionId"]
            doc = cls._get_document(study, doc_id)
            if doc:
                doc["instanceType"] = "StudyDefinitionDocumentVersion"
                # Tweak type for dates
                for date in doc["dateValues"]:
                    if date["type"]["code"] == "C132352":
                        date["type"]["code"] = "C71476"
                        date["type"]["decode"] = "Approval Date"
                    elif date["type"]["code"] == "C207598":
                        date["type"]["code"] = "C215663"
                        date["type"]["decode"] = "Effective Date"
                    else:
                        pass
                items = []
                for content in doc["contents"]:
                    content["displaySectionTitle"] = True
                    content["displaySectionNumber"] = True
                    id = f"{content['id']}_ITEM"
                    items.append(
                        {
                            "id": id,
                            "name": content["name"],
                            "text": content["text"],
                            "instanceType": "NarrativeContentItem",
                        }
                    )
                version["narrativeContentItems"] = items

        # StudyVersion
        for version in study["versions"]:
            version["documentVersionIds"] = [version["documentVersionId"]]
            version.pop("documentVersionId")
            version["referenceIdentifiers"] = []
            version["abbreviations"] = []
            version["roles"] = []
            version["administrableProducts"] = []
            version["notes"] = []

            # Move the organization to a StudyVersion collection
            organizations = []
            for identifier in version["studyIdentifiers"]:
                Convert._move(identifier, "studyIdentifier", "text")
                org = identifier["studyIdentifierScope"]
                organizations.append(org)
                Convert._move(org, "organizationType", "type")
                # print(f"ORG: {org}")
                if org["type"]["code"] == "C70793":
                    org["type"]["code"] = "C54149"
                    org["type"]["decode"] = "Drug Company"
                if org["type"]["code"] == "C93453":
                    org["type"]["decode"] = "Study Registry"
                identifier["scopeId"] = org["id"]
                identifier.pop("studyIdentifierScope")
            version["organizations"] = organizations

            # Tweak type for dates
            for date in version["dateValues"]:
                if date["type"]["code"] == "C132352":
                    date["type"]["code"] = "C71476"
                    date["type"]["decode"] = "Approval Date"
                elif date["type"]["code"] == "C207598":
                    date["type"]["code"] = "C215663"
                    date["type"]["decode"] = "Effective Date"
                else:
                    pass

            # Move the criteria to a StudyVersion collection
            version_ec_items = []
            for index, study_design in enumerate(version["studyDesigns"]):
                criteria = []
                if "population" in study_design:
                    if not study_design["population"]:
                        study_design["population"] = {
                            "id": "Population_Empty",
                            "name": "EMPTY_POPULATION",
                            "includesHealthySubjects": True,
                            "instanceType": "PopulationDefinition",
                        }
                    else:
                        study_design["population"] = Convert._convert_population(
                            study_design["population"], criteria
                        )
                        if study_design["population"] is not None:
                            if "cohorts" in study_design["population"]:
                                new_cohorts = []
                                for cohort in study_design["population"]["cohorts"]:
                                    cohort = Convert._convert_population(
                                        cohort, criteria
                                    )
                                    new_cohorts.append(cohort)
                                study_design["population"]["cohorts"] = new_cohorts

                # print(f"POPULATION: {study_design['population']}")
                ec, ec_items = Convert._split_criteria(criteria)
                version_ec_items += ec_items
                study_design["eligibilityCriteria"] = ec
            version["eligibilityCriterionItems"] = ec_items

            # Update for study designs
            for study_design in version["studyDesigns"]:
                # if "blindingSchema" in study_design:
                #    study_design["blindingSchema"] = Convert._convert_code_to_alias(study_design["blindingSchema"])
                Convert._move(study_design, "trialIntentTypes", "intentTypes")
                Convert._move(study_design, "trialTypes", "subTypes")
                Convert._move(study_design, "interventionModel", "model")
                study_design["studyPhase"] = version["studyPhase"]
                version.pop("studyPhase")
                study_design["studyType"] = version["studyType"]
                version.pop("studyType")
                study_design["instanceType"] = "InterventionalStudyDesign"

            # Process the amendments
            for index, amendment in enumerate(version["amendments"]):
                amendment["name"] = f"AMENDMENT {index + 1}"
                Convert._convert_subject_enrollment(amendment["enrollments"])
                amendment["geographicScopes"] = [
                    {
                        "id": f"{amendment['id']}_GeoScope",
                        "type": {
                            "id": f"{amendment['id']}_GeoScopeType",
                            "code": "C68846",
                            "codeSystem": "cdisc.org",
                            "codeSystemVersion": "2023-12-15",
                            "decode": "Global",
                            "instanceType": "Code",
                        },
                        "instanceType": "GeographicScope",
                    }
                ]

            # Move the main collections
            interventions = []
            dictionaries = []
            conditions = []
            bcs = []
            bc_cat = []
            bc_surrogates = []
            for study_design in version["studyDesigns"]:
                interventions += study_design["studyInterventions"]
                dictionaries += study_design["dictionaries"]
                conditions += study_design["conditions"]
                bcs += study_design["biomedicalConcepts"]
                bc_cat += study_design["bcCategories"]
                bc_surrogates += study_design["bcSurrogates"]
                study_design["studyInterventions"] = [
                    x["id"] for x in study_design["studyInterventions"]
                ]
                study_design.pop("dictionaries")
            for bc in bcs:
                for property in bc["properties"]:
                    for index, code in enumerate(property["responseCodes"]):
                        code["name"] = f"PROPERTY_{index + 1}"
            version["studyInterventions"] = interventions
            version["dictionaries"] = dictionaries
            version["conditions"] = conditions
            version["biomedicalConcepts"] = bcs
            version["bcCategories"] = bc_cat
            version["bcSurrogates"] = bc_surrogates
        return Wrapper.model_validate(wrapper)

    @staticmethod
    def _convert_population(population: dict, criteria: list) -> dict:
        # print(f"POPULATION: {population}")
        if population is None:
            return None
        if "plannedAge" in population:
            population["plannedAge"] = Convert._convert_range(population["plannedAge"])
        if "plannedCompletionNumber" in population:
            population["plannedCompletionNumberRange"] = Convert._convert_range(
                population["plannedCompletionNumber"]
            )
            population.pop("plannedCompletionNumber")
        if "plannedEnrollmentNumber" in population:
            population["plannedEnrollmentNumberRange"] = Convert._convert_range(
                population["plannedEnrollmentNumber"]
            )
            population.pop("plannedEnrollmentNumber")
        if "criteria" in population:
            criteria += population["criteria"]
            population["criterionIds"] = [x["id"] for x in criteria]
            population.pop("criteria")
        return population

    @staticmethod
    def _convert_subject_enrollment(collection: list) -> list:
        for item in collection:
            scope = {
                "id": f"{item['id']}_GeoScope",
                "type": item["type"],
                "code": item["code"],
                "instanceType": "GeographicScope",
            }
            item["forGeographicScope"] = scope
            suffix = (
                f"_{scope['code']['standardCode']['decode']}" if item["code"] else ""
            )
            item["name"] = f"{scope['type']['decode']}{suffix}"
            item.pop("type")
            item.pop("code")
        return collection

    @staticmethod
    def _split_criteria(criteria: list):
        ecis = []
        for criterion in criteria:
            eci = {
                "id": f"{criterion['id']}_item",
                "name": criterion["name"],
                "label": criterion["label"],
                "description": criterion["description"],
                "text": criterion["text"],
                "dictionaryId": criterion["dictionaryId"],
                "instanceType": "EligibilityCriterionItem",
            }
            criterion.pop("text")
            criterion.pop("dictionaryId")
            criterion["criterionItemId"] = eci["id"]
            ecis.append(eci)
        return criteria, ecis

    @staticmethod
    def _get_document(study: dict, doc_id: str):
        # print(f"DOCUMENT BY: {study['documentedBy']}")
        if len(study["documentedBy"]) == 0:
            return None
        if study["documentedBy"]:
            for doc in study["documentedBy"][0]["versions"]:
                if doc["id"] == doc_id:
                    return doc
        return None

    @staticmethod
    def _convert_range(range: dict) -> dict:
        if not range:
            return None
        for item in ["min", "max"]:
            key = f"{item}Value"
            if range["unit"]:
                unit = Convert._convert_code_to_alias(range["unit"])
                unit["id"] = f"{range['unit']['id']}_Unit"
            else:
                unit = None
            range[key] = {
                "id": f"{range['id']}_Min",
                "value": range[key],
                "unit": unit,
                "instanceType": "Quantity",
            }
        range.pop("unit")
        return range

    @staticmethod
    def _convert_code_to_alias(code: dict) -> dict:
        if not code:
            return None
        return {
            "id": f"{code['id']}_AliasCode",
            "standardCode": code,
            "standardAliasCode": [],
            "instanceType": "AliasCode",
        }

    @staticmethod
    def _move(data: dict, from_key: str, to_key: str) -> None:
        data[to_key] = data[from_key]
        data.pop(from_key)



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00013(RuleTemplate):
    """
    DDF00013: If a biomedical concept property is required then it must also be enabled, while if it is not enabled then it must not be required.

    Applies to: BiomedicalConceptProperty
    Attributes: isRequired, isEnabled
    """

    def __init__(self):
        super().__init__(
            "DDF00013",
            RuleTemplate.ERROR,
            "If a biomedical concept property is required then it must also be enabled, while if it is not enabled then it must not be required.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



class Tick:
    def __init__(self, duration: str = None, value: int = None):
        if duration:
            self._tick = self._duration_to_ticks(duration)
        elif value:
            self._tick = value
        else:
            self._tick = 0

    @property
    def tick(self):
        return self._tick

    def __str__(self) -> str:
        intervals = (
            ("weeks", 604800),  # 60 * 60 * 24 * 7
            ("days", 86400),  # 60 * 60 * 24
            ("hours", 3600),  # 60 * 60
            ("minutes", 60),
            ("seconds", 1),
        )
        result = []
        seconds = self._tick
        for name, count in intervals:
            value = seconds // count
            if value:
                seconds -= value * count
                if value == 1:
                    name = name.rstrip("s")
                result.append("{} {}".format(value, name))
        return ", ".join(result)

    def _duration_to_ticks(self, duration: str) -> int:
        if duration.startswith("PT"):
            if duration.endswith("H"):
                return int(duration[2:-1]) * 60 * 60
            elif duration.endswith("M"):
                return int(duration[2:-1]) * 60
            elif duration.endswith("S"):
                return int(duration[2:-1])
            else:
                raise Exception(f"Failed to decode duration '{duration}")
        else:
            if duration.endswith("Y"):
                return int(duration[1:-1]) * 365 * 24 * 60 * 60
            elif duration.endswith("M"):
                return int(duration[1:-1]) * 30 * 24 * 60 * 60
            elif duration.endswith("W"):
                return int(duration[1:-1]) * 7 * 24 * 60 * 60
            elif duration.endswith("D"):
                return int(duration[1:-1]) * 24 * 60 * 60
            else:
                raise Exception(f"Failed to decode duration '{duration}")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00051(RuleTemplate):
    """
    DDF00051: A timing's type must be specified using the Timing Type Value Set Terminology (C201264) DDF codelist.

    Applies to: Timing
    Attributes: type
    """

    def __init__(self):
        super().__init__(
            "DDF00051",
            RuleTemplate.ERROR,
            "A timing's type must be specified using the Timing Type Value Set Terminology (C201264) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00037(RuleTemplate):
    """
    DDF00037: At least one scheduled activity instance within a timeline must point to a timeline exit.

    Applies to: ScheduledActivityInstance
    Attributes: timelineExit
    """

    def __init__(self):
        super().__init__(
            "DDF00037",
            RuleTemplate.ERROR,
            "At least one scheduled activity instance within a timeline must point to a timeline exit.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Union
from .api_base_model import ApiBaseModel
from .study import Study
from .study_version import StudyVersion
from .study_design import StudyDesign
from .study_definition_document_version import StudyDefinitionDocumentVersion


class Wrapper(ApiBaseModel):
    study: Study
    usdmVersion: str
    systemName: Union[str, None] = None
    systemVersion: Union[str, None] = None

    def first_version(self) -> StudyVersion | None:
        return self.study.first_version() if self.study else None

    def study_version_and_design(
        self, id: str
    ) -> tuple[Study | None, StudyVersion | None, StudyDesign | None]:
        study: Study = self.study
        if study:
            version: StudyVersion = study.first_version()
            if version:
                return study, version, version.find_study_design(id)
        return None, None, None

    def to_html(self, template: str) -> str:
        study: Study = self.study
        version: StudyVersion = self.first_version()
        return (
            version.to_html(template, study.document_map()) if version and study else ""
        )

    def study_document_version(
        self, template: str
    ) -> StudyDefinitionDocumentVersion | None:
        study: Study = self.study
        version: StudyVersion = self.first_version()
        return (
            version.study_document_version(template, study.document_map())
            if version and study
            else None
        )



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0276(RuleTemplate):
    """
    CHK0276: An observational study (including patient registries) is expected to have a study phase decode value of \"NOT APPLICABLE\".

    Applies to: ObservationalStudyDesign
    Attributes: studyPhase
    """

    def __init__(self):
        super().__init__(
            "CHK0276",
            RuleTemplate.ERROR,
            'An observational study (including patient registries) is expected to have a study phase decode value of "NOT APPLICABLE".',
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0242(RuleTemplate):
    """
    CHK0242: Each study definition document version is expected to be referenced by either a study version or a study design.

    Applies to: StudyVersion, ObservationalStudyDesign, InterventionalStudyDesign
    Attributes: documentVersions
    """

    def __init__(self):
        super().__init__(
            "CHK0242",
            RuleTemplate.ERROR,
            "Each study definition document version is expected to be referenced by either a study version or a study design.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0174(RuleTemplate):
    """
    CHK0174: An eligility criterion must not be referenced by both a study design population and any of the cohorts of the same study design population.

    Applies to: StudyVersion
    Attributes: criteria
    """

    def __init__(self):
        super().__init__(
            "CHK0174",
            RuleTemplate.ERROR,
            "An eligility criterion must not be referenced by both a study design population and any of the cohorts of the same study design population.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithId


class StudyCell(ApiBaseModelWithId):
    armId: str
    epochId: str
    elementIds: List[str]
    instanceType: Literal["StudyCell"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0220(RuleTemplate):
    """
    CHK0220: If a dose is specified, then a corresponding administrable product must also be specified either directly or embedded in the medical device and vice versa.

    Applies to: Administration
    Attributes: dose, administrableProduct
    """

    def __init__(self):
        super().__init__(
            "CHK0220",
            RuleTemplate.ERROR,
            "If a dose is specified, then a corresponding administrable product must also be specified either directly or embedded in the medical device and vice versa.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal, Union, List
from .api_base_model import ApiBaseModelWithId
from .code import Code


class Address(ApiBaseModelWithId):
    text: Union[str, None] = None
    lines: List[str] = []
    city: Union[str, None] = None
    district: Union[str, None] = None
    state: Union[str, None] = None
    postalCode: Union[str, None] = None
    country: Union[Code, None] = None
    instanceType: Literal["Address"]

    def set_text(self) -> None:
        text = ""
        for line in self.lines:
            text = Address._concat(text, line)
        for attr in ["city", "district", "state", "postalCode"]:
            text = Address._concat(text, self.__getattribute__(attr))
        self.text = Address._concat(text, self.country.decode) if self.country else text

    @staticmethod
    def _concat(text, value):
        if text:
            return text + ", " + value if value else text
        else:
            return value if value else ""



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00058(RuleTemplate):
    """
    DDF00058: Within an indication, if more indication codes are defined, they must be distinct.

    Applies to: Indication
    Attributes: codes
    """

    def __init__(self):
        super().__init__(
            "DDF00058",
            RuleTemplate.ERROR,
            "Within an indication, if more indication codes are defined, they must be distinct.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00043(RuleTemplate):
    """
    DDF00043: A unit must not be specified for a planned enrollment number or a planned completion number.

    Applies to: StudyDesignPopulation, StudyCohort
    Attributes: plannedEnrollmentNumber, plannedCompletionNumber
    """

    def __init__(self):
        super().__init__(
            "DDF00043",
            RuleTemplate.ERROR,
            "A unit must not be specified for a planned enrollment number or a planned completion number.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00102(RuleTemplate):
    """
    DDF00102: A scheduled activity instance must only reference a timeline exit that is defined within the same schedule timeline as the scheduled activity instance.

    Applies to: ScheduledActivityInstance
    Attributes: timelineExit
    """

    def __init__(self):
        super().__init__(
            "DDF00102",
            RuleTemplate.ERROR,
            "A scheduled activity instance must only reference a timeline exit that is defined within the same schedule timeline as the scheduled activity instance.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm4.api.api_base_model import ApiBaseModelWithId


class PathError(Exception):
    pass


class DuplicateError(Exception):
    pass


class CrossReference:
    def __init__(self):
        self._by_name = {}
        self._by_id = {}

    def clear(self):
        self._by_name = {}
        self._by_id = {}

    def add(self, object: ApiBaseModelWithId, name: str) -> None:
        name = object.name if hasattr(object, "name") else name
        if name:
            self._add_to_collection(name, self._by_name, object)
        self._add_to_collection(object.id, self._by_id, object)

    def _add_to_collection(self, key_text: str, collection: dict, object):
        if key_text:
            klass = object.__class__
            key = self._key(klass, key_text)
            if key not in collection:
                collection[key] = object
            else:
                raise DuplicateError(
                    f"Duplicate cross reference detected, klass='{self._klass_name(klass)}', key='{key_text}'"
                )

    def get_by_name(self, klass, name: str) -> ApiBaseModelWithId:
        return self._get(klass, name, self._by_name)

    def get_by_id(self, klass, id: str) -> ApiBaseModelWithId:
        return self._get(klass, id, self._by_id)

    def _get(self, klass, key_text: str, collection: dict) -> ApiBaseModelWithId:
        key = self._key(klass, key_text)
        if key in collection:
            return collection[key]
        else:
            return None

    def get_by_path(self, klass, name, path):
        instance = self.get_by_name(klass, name)
        if instance:
            parts = path.split("/")
            attribute = parts[0].replace("@", "")
            if len(parts) == 1:
                return instance, attribute
            elif len(parts) % 2 == 1:
                for index in range(1, len(parts), 2):
                    try:
                        instance = getattr(instance, attribute)
                    except AttributeError:
                        raise PathError(
                            f"Failed to translate reference path '{path}', attribute '{attribute}' was not found"
                        )
                    attribute = parts[index + 1].replace("@", "")
                    if not parts[index] == instance.__class__.__name__:
                        raise PathError(
                            f"Failed to translate reference path '{path}', class mismtach, expecting '{parts[index]}', found '{instance.__class__.__name__}'"
                        )
                if instance and attribute:
                    if not self.get_by_id(instance.__class__, instance.id):
                        self.add(instance.id, instance)
                    return instance, attribute
                else:
                    raise PathError(
                        f"Failed to translate reference path '{path}', path was not found"
                    )
            else:
                raise PathError(
                    f"Failed to translate reference path '{path}', format error"
                )
        else:
            raise PathError(
                f"Failed to translate reference path '{path}', could not find start instance '{klass}', '{name}'"
            )

    def _key(self, klass, text: str) -> str:
        klass_name = self._klass_name(klass)
        return f"{klass_name}.{text}"

    def _klass_name(self, klass):
        return klass if isinstance(klass, str) else klass.__name__



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0239(RuleTemplate):
    """
    CHK0239: Each study enrollment must apply to either a geographic scope, a study site, or a study cohort.

    Applies to: SubjectEnrollment
    Attributes: appliesTo
    """

    def __init__(self):
        super().__init__(
            "CHK0239",
            RuleTemplate.ERROR,
            "Each study enrollment must apply to either a geographic scope, a study site, or a study cohort.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .alias_code import AliasCode
from .api_base_model import ApiBaseModelWithIdNameAndLabel
from .response_code import ResponseCode
from .comment_annotation import CommentAnnotation


class BiomedicalConceptProperty(ApiBaseModelWithIdNameAndLabel):
    isRequired: bool
    isEnabled: bool
    datatype: str
    responseCodes: List[ResponseCode] = []
    code: AliasCode
    notes: List[CommentAnnotation] = []
    instanceType: Literal["BiomedicalConceptProperty"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00026(RuleTemplate):
    """
    DDF00026: A scheduled activity instance must not point (via the \"timeline\" relationship) to the timeline in which it is specified.

    Applies to: ScheduledActivityInstance
    Attributes: timeline
    """

    def __init__(self):
        super().__init__(
            "DDF00026",
            RuleTemplate.ERROR,
            'A scheduled activity instance must not point (via the "timeline" relationship) to the timeline in which it is specified.',
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Literal
from .api_base_model import ApiBaseModelWithId
from .substance import Substance
from .code import Code


class Ingredient(ApiBaseModelWithId):
    role: Code
    substance: Substance
    instanceType: Literal["Ingredient"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00052(RuleTemplate):
    """
    DDF00052: All standard code aliases referenced by an instance of the alias code class must be unique.

    Applies to: AliasCode
    Attributes: standardCodeAliases
    """

    def __init__(self):
        super().__init__(
            "DDF00052",
            RuleTemplate.ERROR,
            "All standard code aliases referenced by an instance of the alias code class must be unique.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00022(RuleTemplate):
    """
    DDF00022: An instance of a class must not refer to itself as its next instance.

    Applies to: StudyEpoch, Encounter, Activity, NarrativeContent, EligibilityCriterion
    Attributes: next
    """

    def __init__(self):
        super().__init__(
            "DDF00022",
            RuleTemplate.ERROR,
            "An instance of a class must not refer to itself as its next instance.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00091(RuleTemplate):
    """
    DDF00091: When a condition applies to a procedure, activity, biomedical concept, biomedical concept category, or biomedical concept surrogate then an instance must be available in the corresponding class with the specified id.

    Applies to: Condition
    Attributes: appliesTo
    """

    def __init__(self):
        super().__init__(
            "DDF00091",
            RuleTemplate.ERROR,
            "When a condition applies to a procedure, activity, biomedical concept, biomedical concept category, or biomedical concept surrogate then an instance must be available in the corresponding class with the specified id.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code


class CommentAnnotation(ApiBaseModelWithId):
    text: str
    codes: List[Code] = []
    instanceType: Literal["CommentAnnotation"]



import re
from uuid import uuid4
from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.assembler.base_assembler import BaseAssembler
from usdm4.assembler.encoder import Encoder
from usdm4.assembler.identification_assembler import IdentificationAssembler
from usdm4.assembler.study_design_assembler import StudyDesignAssembler
from usdm4.assembler.document_assembler import DocumentAssembler
from usdm4.assembler.population_assembler import PopulationAssembler
from usdm4.assembler.amendments_assembler import AmendmentsAssembler
from usdm4.assembler.timeline_assembler import TimelineAssembler
from usdm4.builder.builder import Builder
from usdm4.api.study import Study
from usdm4.api.study_version import (
    StudyVersion,
    CS_EXT_URL,
    OV_EXT_URL,
    CC_EXT_URL,
    CN_EXT_URL,
    ME_EXT_URL,
    SS_EXT_URL,
    SAL_EXT_URL,
)
from usdm4.api.geographic_scope import GeographicScope
from usdm4.api.governance_date import GovernanceDate
from usdm4.api.extension import ExtensionAttribute


class StudyAssembler(BaseAssembler):
    """
    Assembler responsible for creating the top-level Study object and its associated StudyVersion.

    This assembler coordinates data from multiple other assemblers to create the complete study
    structure, including study metadata, versions, governance dates, and cross-references to
    study designs, documents, and identification information.
    """

    MODULE = "usdm4.assembler.study_assembler.StudyAssembler"

    def __init__(self, builder: Builder, errors: Errors):
        """
        Initialize the StudyAssembler.

        Args:
            builder (Builder): The builder instance for creating USDM objects
            errors (Errors): Error handling instance for logging issues
        """
        super().__init__(builder, errors)
        self._encoder = Encoder(builder, errors)
        self.clear()

    def clear(self):
        self._study = None
        self._dates = []

    def execute(
        self,
        data: dict,
        identification_assembler: IdentificationAssembler,
        study_design_assembler: StudyDesignAssembler,
        document_assembler: DocumentAssembler,
        population_assembler: PopulationAssembler,
        amendments_assembler: AmendmentsAssembler,
        timeline_assembler: TimelineAssembler,
    ) -> None:
        """
        Creates the top-level Study object and its associated StudyVersion from study data.

        Args:
            data (dict): A dictionary containing core study information.
                        The data parameter must have the following structure:

                        {
                            "name": str,               # Internal name/identifier for the study
                            "label": str,              # Human-readable display label for the study
                            "version": str,            # Version identifier for this study version
                            "rationale": str,          # Rationale or reason for this study version
                            # Additional optional fields may include:
                            # "description": str,       # Detailed study description
                            # "objectives": list,       # List of study objectives
                            # "indications": list,      # List of study indications
                            # "interventions": list,    # List of study interventions
                            # "sponsor_approval_date": str,  # Date of sponsor approval (ISO format)
                        }

                        Required fields:
                        - "name": Internal identifier/name for the study
                        - "label": Display label for the study
                        - "version": Version identifier (e.g., "1.0", "2.1", etc.)
                        - "rationale": Explanation for this version of the study

            identification_assembler (IdentificationAssembler): Assembler containing study
                identifiers, titles, and organization information
            study_design_assembler (StudyDesignAssembler): Assembler containing the complete
                study design structure (arms, epochs, activities, etc.)
            document_assembler (DocumentAssembler): Assembler containing protocol documents
                and governance dates

        Returns:
            None: The created study is stored in self._study property

        Note:
            This method creates both a StudyVersion object (containing version-specific data)
            and a top-level Study object (containing the study container and metadata).
            The StudyVersion aggregates data from all other assemblers, while the Study
            object serves as the root container.
        """
        try:
            # Create the dates
            valid_date = self._create_date(data)

            # Extensions
            extensions = []
            if not valid_date and "sponsor_approval_date" in data:
                # Create where approval infomration is extension
                self._create_extension(
                    extensions, SAL_EXT_URL, data["sponsor_approval_date"]
                )
            if "confidentiality" in data:
                # Create confidentiality extension
                self._create_extension(extensions, CS_EXT_URL, data["confidentiality"])
            if "original_protocol" in data:
                # Create original protocol
                extensions.append(
                    self._builder.create(
                        ExtensionAttribute,
                        {
                            "url": OV_EXT_URL,
                            "valueBoolean": self._encoder.to_boolean(
                                data["original_protocol"]
                            ),
                        },
                    )
                )
            if identification_assembler.compound_codes:
                # Compound codes
                self._create_extension(
                    extensions, CC_EXT_URL, identification_assembler.compound_codes
                )
            if identification_assembler.compound_names:
                # compound names
                self._create_extension(
                    extensions, CN_EXT_URL, identification_assembler.compound_names
                )
            if identification_assembler.sponsor_signatory:
                # Sponsor signatory
                self._create_extension(
                    extensions, SS_EXT_URL, identification_assembler.sponsor_signatory
                )
            if identification_assembler.medical_expert:
                # Medical expert
                self._create_extension(
                    extensions, ME_EXT_URL, identification_assembler.medical_expert
                )

            # Create StudyVersion parameters by combining data from all assemblers
            params = {
                "versionIdentifier": data["version"],  # Version ID from input data
                "rationale": data["rationale"],  # Version rationale from input data
                "titles": identification_assembler.titles,  # Study titles from identification
                "dateValues": self._dates
                + document_assembler.dates,  # Combined governance dates
                "studyDesigns": [
                    study_design_assembler.study_design
                ],  # Study design structure
                "documentVersionIds": [document_assembler.document_version.id]
                if document_assembler.document_version
                else [],  # Document references
                "studyIdentifiers": identification_assembler.identifiers,  # Study identifiers
                "organizations": identification_assembler.organizations,  # Sponsor/organization info
                "eligibilityCriterionItems": population_assembler.criteria_items,
                "narrativeContentItems": document_assembler.contents,
                "amendments": [amendments_assembler.amendment]
                if amendments_assembler.amendment
                else [],
                "conditions": timeline_assembler.conditions,
                "biomedicalConcepts": timeline_assembler.biomedical_concepts,
                "bcSurrogates": timeline_assembler.biomedical_concept_surrogates,
                "roles": identification_assembler.roles,
                "extensionAttributes": extensions,
            }
            study_version = self._builder.create(StudyVersion, params)

            # Create the top-level Study container object
            study_name, study_label = self._get_study_name_label(data["name"])
            self._study = self._builder.create(
                Study,
                {
                    "id": uuid4(),  # Generate unique study ID
                    "name": study_name,  # Internal study name
                    "label": study_label,  # Display study label
                    "description": "The top-level study container",  # Default description
                    "versions": [study_version],  # Include the created version
                    "documentedBy": [document_assembler.document]
                    if document_assembler.document
                    else [],  # Reference to protocol document
                },
            )
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception("Failed during creation of study", e, location)

    @property
    def study(self) -> Study:
        return self._study

    def _create_extension(
        self, extensions: list[ExtensionAttribute], url: str, value: str
    ) -> None:
        try:
            extensions.append(
                self._builder.create(
                    ExtensionAttribute, {"url": url, "valueString": value}
                )
            )
        except Exception as e:
            self._errors.exception(
                "Failed during creation of extesnion",
                e,
                KlassMethodLocation(self.MODULE, "_create_extension"),
            )

    def _create_date(self, data: dict) -> bool:
        result = False
        try:
            if actual_date := self._encoder.to_date(data["sponsor_approval_date"]):
                sponsor_approval_date_code = self._builder.cdisc_code(
                    "C71476", "Approval Date"
                )
                global_code = self._builder.cdisc_code("C68846", "Global")
                global_scope = self._builder.create(
                    GeographicScope, {"type": global_code}
                )
                approval_date = self._builder.create(
                    GovernanceDate,
                    {
                        "name": "SPONSOR-APPORVAL-DATE",  # Human-readable name
                        "type": sponsor_approval_date_code,  # CDISC code for date type
                        "dateValue": actual_date,
                        "geographicScopes": [global_scope],  # Global scope application
                    },
                )
                if approval_date:
                    self._dates.append(approval_date)
                    result = True
            else:
                self._errors.warning(
                    "No sponsor approval date detected",
                    KlassMethodLocation(self.MODULE, "_create_date"),
                )
            return result
        except Exception as e:
            self._errors.exception(
                "Failed during creation of governance date",
                e,
                KlassMethodLocation(self.MODULE, "_create_date"),
            )
            return False

    def _get_study_name_label(self, options: dict) -> tuple[str, str]:
        items = ["identifier", "acronym", "compound"]
        for item in items:
            if item in options and options[item]:
                name = re.sub(r"[\W_]+", "", options[item].upper())
                self._errors.info(
                    f"Study name set to '{name}'",
                    location=KlassMethodLocation(self.MODULE, "_get_study_name_label"),
                )
                return name, options[item]
        return "", ""



from usdm4.api.study_design import StudyDesign
from usdm4.api.schedule_timeline import ScheduleTimeline
from usdm4.api.scheduled_instance import ScheduledActivityInstance
from usdm4.api.code import Code
from usdm4.api.timing import Timing
from .tick import Tick
from simple_error_log import Errors


class Timepoint:
    def __init__(
        self,
        study_design: StudyDesign,
        timeline: ScheduleTimeline,
        sai: ScheduledActivityInstance,
        errors: Errors,
        id: int,
        offset: int,
    ):
        self._sai: ScheduledActivityInstance = sai
        self._errors = errors
        self._timeline = timeline
        self._study_design = study_design
        self._tick: int = self._calculate_hop(timeline, sai) + offset
        self._id: str = f"TP_{id}"
        self._edges: list[str] = []

    @property
    def tick(self) -> int:
        return self._tick

    @property
    def id(self) -> str:
        return self._id

    @property
    def activities(self) -> bool:
        return len(self._sai.activityIds) > 0

    def add_edge(self, next: "Timepoint") -> None:
        self._edges.append(next.id)

    def activity_timelines(self) -> list[ScheduleTimeline]:
        activities = [
            self._study_design.find_activity(x) for x in self._sai.activityIds
        ]
        return [
            self._study_design.find_timeline(x.timelineId)
            for x in activities
            if x.timelineId
        ]

    def to_dict(self):
        parents = self._study_design.activity_parent()
        activities = [
            self._study_design.find_activity(x) for x in self._sai.activityIds
        ]
        return {
            "id": self._id,
            "tick": self._tick,
            "time": str(Tick(value=self._tick)),
            "label": self._sai.label,
            "encounter": self._study_design.find_encounter(self._sai.encounterId).label
            if self._sai.encounterId
            else None,
            "main_timeline": self._timeline.mainTimeline,
            "activities": {
                "items": [
                    {
                        "label": x.label,
                        "parent": self._study_design.find_activity(parents[x.id]).label
                        if x.id in parents
                        else None,
                        "procedures": [p.label for p in x.definedProcedures],
                        "notes": [
                            {
                                "text": x.text,
                                "codes": [self._code_dict(c) for c in x.codes],
                            }
                            for x in x.notes
                        ],
                    }
                    for x in activities
                ]
            },
            "edges": self._edges,
        }

    def _code_dict(self, code: Code) -> dict:
        return {
            "code": code.code,
            "decode": code.decode,
            "system": code.codeSystem,
            "version": code.codeSystemVersion,
        }

    def _calculate_hop(
        self, timeline: ScheduleTimeline, sai: ScheduledActivityInstance
    ) -> int:
        return self._calculate_next_hop(timeline, sai, 0)

    def _calculate_next_hop(
        self, timeline: ScheduleTimeline, sai: ScheduledActivityInstance, tick: int
    ) -> int:
        # print(f"NEXT HOP: {timeline.id}, {sai.id}")
        timing: Timing = timeline.find_timing_from(sai.id)
        # print(f"TIMING: {timing}")
        to_sai = timeline.find_timepoint(timing.relativeToScheduledInstanceId)
        before = timing.type.code == "C201357"
        to_timing: Timing = timeline.find_timing_from(to_sai.id)
        value = self._calculate_tick(timing)
        new_tick = tick - value if before else tick + value
        if to_timing.type.code == "C201358":  # Anchor, so stop
            return new_tick
        else:
            return self._calculate_next_hop(timeline, to_sai, new_tick)

    def _calculate_tick(self, timing: Timing) -> int:
        try:
            return Tick(duration=timing.value).tick
        except Exception as e:
            self._errors.error(f"Failed to decode duration '{timing.value}', {e}")
            return 0



from typing import List, Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .comment_annotation import CommentAnnotation


class SyntaxTemplate(ApiBaseModelWithIdNameLabelAndDesc):
    text: str
    dictionaryId: Union[str, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["SyntaxTemplate"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00107(RuleTemplate):
    """
    DDF00107: A scheduled activity instance must only have a sub-timeline that is defined within the same study design as the scheduled activity instance.

    Applies to: ScheduledActivityInstance
    Attributes: timeline
    """

    def __init__(self):
        super().__init__(
            "DDF00107",
            RuleTemplate.ERROR,
            "A scheduled activity instance must only have a sub-timeline that is defined within the same study design as the scheduled activity instance.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0183(RuleTemplate):
    """
    CHK0183: If a section number is to be displayed then a number must be specified and vice versa.

    Applies to: NarrativeContent
    Attributes: sectionNumber, displaySectionNumber
    """

    def __init__(self):
        super().__init__(
            "CHK0183",
            RuleTemplate.ERROR,
            "If a section number is to be displayed then a number must be specified and vice versa.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0189(RuleTemplate):
    """
    CHK0189: A study definition document version's status must be specifed using the status Value Set Terminology (C188723) DDF codelist.

    Applies to: StudyDefinitionDocumentVersion
    Attributes: status
    """

    def __init__(self):
        super().__init__(
            "CHK0189",
            RuleTemplate.ERROR,
            "A study definition document version's status must be specifed using the status Value Set Terminology (C188723) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .api_base_model import ApiBaseModelWithId
from .code import Code


class AliasCode(ApiBaseModelWithId):
    standardCode: Code
    standardCodeAliases: List[Code] = []
    instanceType: Literal["AliasCode"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00035(RuleTemplate):
    """
    DDF00035: Within a code system and corresponding version, a one-to-one relationship between code and decode is expected.

    Applies to: Code
    Attributes: code, decode
    """

    def __init__(self):
        super().__init__(
            "DDF00035",
            RuleTemplate.ERROR,
            "Within a code system and corresponding version, a one-to-one relationship between code and decode is expected.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00005(RuleTemplate):
    """
    DDF00005: Every study version must have exactly one study identifier with an identifier scope that references a clinical study sponsor organization.

    Applies to: StudyIdentifier
    Attributes: studyIdentifierScope
    """

    def __init__(self):
        super().__init__(
            "DDF00005",
            RuleTemplate.ERROR,
            "Every study version must have exactly one study identifier with an identifier scope that references a clinical study sponsor organization.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



"""
Assembler for creating StudyAmendment objects from input data.

This module processes amendment data and constructs USDM4-compliant StudyAmendment
objects including reasons, enrollment information, and geographic scopes.
"""

import re
from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.assembler.base_assembler import BaseAssembler
from usdm4.assembler.encoder import Encoder
from usdm4.assembler.document_assembler import DocumentAssembler
from usdm4.builder.builder import Builder
from usdm4.api.quantity_range import Quantity
from usdm4.api.geographic_scope import GeographicScope
from usdm4.api.subject_enrollment import SubjectEnrollment
from usdm4.api.study_amendment_reason import StudyAmendmentReason
from usdm4.api.study_amendment import StudyAmendment, SI_EXT_URL
from usdm4.api.study_amendment_impact import StudyAmendmentImpact
from usdm4.api.study_change import StudyChange
from usdm4.api.document_content_reference import DocumentContentReference
from usdm4.api.code import Code
from usdm4.api.extension import ExtensionAttribute


class AmendmentsAssembler(BaseAssembler):
    """
    Assembler that creates StudyAmendment objects from structured input data.

    Handles the creation of amendments including their reasons, enrollment data,
    and geographic scope information. Supports global, country-specific, and
    region-specific scopes.
    """

    MODULE = "usdm4.assembler.amendments_assembler.AmenementsAssembler"

    def __init__(self, builder: Builder, errors: Errors):
        """
        Initialize the AmendmentsAssembler.

        Args:
            builder: The Builder instance for creating USDM4 objects.
            errors: The Errors instance for logging errors and information.
        """
        super().__init__(builder, errors)
        self._encoder = Encoder(builder, errors)
        self.clear()

    def clear(self):
        """Reset the assembler state by clearing the current amendment."""
        self._amendment = None

    def execute(self, data: dict, document_assembler: DocumentAssembler) -> None:
        """
        Execute the amendment assembly process.

        Processes the input data dictionary and creates a StudyAmendment object.
        Any exceptions during processing are caught and logged.

        Args:
            data: Dictionary containing amendment data with keys like 'identifier',
                  'summary', 'reasons', 'impact', 'enrollment', and 'scope'.
        """
        try:
            self._document_assembler = document_assembler
            if data:
                self._amendment = self._create_amendment(data)
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception("Failed during creation of amendments", e, location)

    @property
    def amendment(self) -> StudyAmendment:
        """
        Get the created StudyAmendment object.

        Returns:
            The StudyAmendment object created by the last execute() call,
            or None if no amendment was created.
        """
        return self._amendment

    def _create_amendment(self, data: dict) -> StudyAmendment:
        """
        Create a StudyAmendment from the provided data.

        Constructs the amendment with primary and secondary reasons, impact flags,
        enrollment information, and geographic scopes.

        Args:
            data: Dictionary containing amendment details.

        Returns:
            A StudyAmendment object, or None if creation fails.
        """
        try:
            self._errors.info(f"Amendment assembler source data {data}")
            reasons = self._create_primary_secondary_reasons(data)
            geo_scopes, site_scopes = self._create_scopes(data)
            params = {
                "name": "AMENDMENT 1",
                "number": data["identifier"],
                "summary": data["summary"],
                "impacts": self._create_amendment_impact(data),
                "primaryReason": reasons["primary"],
                "secondaryReasons": [reasons["secondary"]],
                "enrollments": [self._create_enrollment(data)],
                "geographicScopes": geo_scopes,
                "changes": self._create_changes(data),
                "extensionAttributes": [site_scopes] if site_scopes else [],
            }
            return self._builder.create(StudyAmendment, params)
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "execute")
            self._errors.exception("Failed during creation of amendments", e, location)
            return None

    def _create_primary_secondary_reasons(
        self, data: dict
    ) -> dict[str, StudyAmendmentReason]:
        reasons = {}
        for k, item in data["reasons"].items():
            a_reason = self._encoder.amendment_reason(item)
            a_reason["otherReason"] = a_reason.pop("other_reason")
            reasons[k] = self._builder.create(StudyAmendmentReason, a_reason)
        return reasons

    def _create_changes(self, data: dict) -> list[StudyChange]:
        results = []
        for index, item in enumerate(data["changes"]):
            refs = self._extract_section_numer_and_title(item["section"])
            params = {
                "name": f"CHANGE_{index + 1}",
                "summary": item["description"].strip(),
                "rationale": item["rationale"].strip(),
                "changedSections": refs,
            }
            change = self._builder.create(StudyChange, params)
            if change:
                results.append(change)
        return results

    def _extract_section_numer_and_title(self, text) -> list[DocumentContentReference]:
        results = []
        pattern = r"^(?:Section\s+)?(\d+(?:\.\d+)*),?\s*(.*)$"
        for line in text.strip().split("\n"):
            match = re.match(pattern, line.strip())
            if match:
                params = {
                    "sectionNumber": match.group(1),
                    "sectionTitle": match.group(2),
                    "appliesToId": self._document_assembler.document.id,
                }
                ref = self._builder.create(DocumentContentReference, params)
                if ref:
                    results.append(ref)
                    self._errors.info(
                        f"Extracted section ref from '{line}' -> {params}",
                        KlassMethodLocation(
                            self.MODULE, "_extract_section_numer_and_title"
                        ),
                    )
            else:
                self._errors.error(
                    f"Failed to extract section ref from '{line}'",
                    KlassMethodLocation(
                        self.MODULE, "_extract_section_numer_and_title"
                    ),
                )
        return results

    def _create_amendment_impact(self, data: dict) -> list[StudyAmendmentImpact]:
        try:
            results = []
            self._errors.info(
                f"Creating amendment impacts using {data['impact']}",
                KlassMethodLocation(self.MODULE, "_create_amendment_impact"),
            )
            impact = data["impact"]
            s_and_r = impact["safety_and_rights"]
            r_and_r = impact["reliability_and_robustness"]
            # print(f"\nR-R1: {r_and_r}")
            # print(f"R-R2: {r_and_r["reliability"]}")
            self._create_impact(
                results,
                "C215665",
                "Study Subject Safety",
                s_and_r["safety"]["substantial"],
                s_and_r["safety"]["reason"],
            )
            self._create_impact(
                results,
                "C215666",
                "Study Subject Rights",
                s_and_r["rights"]["substantial"],
                s_and_r["rights"]["reason"],
            )
            self._create_impact(
                results,
                "C215667",
                "Study Data Reliability",
                r_and_r["reliability"]["substantial"],
                r_and_r["reliability"]["reason"],
            )
            self._create_impact(
                results,
                "C215668",
                "Study Data Robustness",
                r_and_r["robustness"]["substantial"],
                r_and_r["robustness"]["reason"],
            )
            return results
        except Exception as e:
            self._errors.exception(
                "Failed during creation of amendment impacts",
                e,
                KlassMethodLocation(self.MODULE, "_create_amendment_impact"),
            )
            return []

    def _create_impact(
        self,
        results: list[StudyAmendmentImpact],
        code: str,
        decode: str,
        is_substantial: bool,
        text: str,
    ) -> None:
        type_code = self._builder.cdisc_code(code, decode)
        params = {
            "text": text,
            "isSubstantial": is_substantial,
            "type": type_code,
        }
        item = self._builder.create(StudyAmendmentImpact, params)
        if item:
            results.append(item)

    def _create_enrollment(self, data: dict) -> SubjectEnrollment:
        """
        Create a SubjectEnrollment object from the provided data.

        If enrollment data is present, creates a quantity with the specified value
        and unit (percentage if '%'). Otherwise, creates a default enrollment with
        value 0.

        Args:
            data: Dictionary that may contain an 'enrollment' key with 'value' and 'unit'.

        Returns:
            A SubjectEnrollment object, or None if creation fails.
        """
        try:
            global_code = self._builder.cdisc_code("C68846", "Global")
            params = {
                "type": global_code,
                "code": None,
            }
            geo_scope = self._builder.create(GeographicScope, params)
            if "enrollment" in data:
                unit_alias = None
                if data["enrollment"]["unit"] == "%":
                    unit_code = self._builder.cdisc_code("C25613", "Percentage")
                    unit_alias = (
                        self._builder.alias_code(unit_code) if unit_code else None
                    )
                quantity = self._builder.create(
                    Quantity,
                    {
                        "value": self._to_int(data["enrollment"]["value"]),
                        "unit": unit_alias,
                    },
                )
                params = {
                    "name": "ENROLLMENT",
                    "forGeographicScope": geo_scope,
                    "quantity": quantity,
                }
            else:
                quantity = self._builder.create(Quantity, {"value": 0, "unit": None})
                params = {
                    "name": "ENROLLMENT",
                    "forGeographicScope": geo_scope,
                    "quantity": quantity,
                }
            return self._builder.create(SubjectEnrollment, params)
        except Exception as e:
            location = KlassMethodLocation(self.MODULE, "_create_enrollment")
            self._errors.exception("Failed during creation of enrollments", e, location)
            return None

    def _to_int(self, item: str | int) -> int:
        try:
            return item if isinstance(item, int) else int(str(item))
        except Exception as e:
            self._errors.exception(
                f"Failed to convert '{item}' to integer value",
                e,
                KlassMethodLocation(self.MODULE, "_to_int"),
            )
            return 0

    def _create_scopes(
        self, data: dict
    ) -> tuple[list[GeographicScope], ExtensionAttribute]:
        extension = None
        results = []
        extensions = []
        if "scope" in data and data["scope"]:
            scope = data["scope"]
            # {"global": True/False, "countries": [], "regions": [], "sites": [], "unknown": []}
            if scope["global"]:
                self._global_scope(results)
            else:
                for part in scope["unknown"]:
                    text = part.strip()
                    if not text:
                        continue
                    # Try to find as a country code first
                    code, decode = self._builder.iso3166_library.code_or_decode(text)
                    if code:
                        country_code = self._encoder.geographic_scope("COUNTRY")
                        self._create_scope(results, country_code, code, decode)
                    else:
                        # If not a country, try as a region code
                        code, decode = self._builder.iso3166_library.region_code(text)
                        if code:
                            region_code = self._encoder.geographic_scope("REGION")
                            self._create_scope(results, region_code, code, decode)
                        else:
                            self._create_site_scope(extensions, text)
                for part in scope["countries"]:
                    code, decode = self._builder.iso3166_library.code_or_decode(part)
                    if code:
                        country_code = self._encoder.geographic_scope("COUNTRY")
                        self._create_scope(results, country_code, code, decode)
                for part in scope["regions"]:
                    code, decode = self._builder.iso3166_library.region_code(part)
                    if code:
                        region_code = self._encoder.geographic_scope("REGION")
                        self._create_scope(results, region_code, code, decode)
                for part in scope["sites"]:
                    self._create_site_scope(extensions, part)
        else:
            # Empty scope string - default to global and log error
            self._global_scope(results)
            self._errors.error(
                "Empty scope found",
                KlassMethodLocation(self.MODULE, "_create_scopes"),
            )
        if extensions:
            extension = self._builder.create(
                ExtensionAttribute,
                {
                    "url": SI_EXT_URL,
                    "extensionAttributes": extensions,
                },
            )
        self._errors.info(
            f"Scopes created in assembler, geo: {results}, sites: {extension}"
        )
        return results, extension

    def _global_scope(self, results: list[GeographicScope]) -> None:
        """
        Add a global geographic scope to the results list.

        Args:
            results: List to append the global scope to.
        """
        global_code = self._encoder.geographic_scope("GLOBAL")
        self._create_scope(results, global_code)

    def _create_site_scope(self, extensions: list, text: str):
        extension = self._builder.create(
            ExtensionAttribute,
            {
                "url": "site-identifier",
                "valueString": text,
            },
        )
        if extension:
            extensions.append(extension)

    def _create_scope(
        self,
        results: list[GeographicScope],
        type: Code,
        code: str = None,
        decode: str = None,
    ) -> None:
        """
        Create a GeographicScope and add it to the results list.

        For country scopes (C25464), looks up the ISO3166 country code.
        For region scopes, looks up the ISO3166 region code.
        Creates an alias code from the standard code if found.

        Args:
            results: List to append the created scope to.
            type: The Code object indicating the scope type (Country, Global, Region).
            code: Optional ISO code string (e.g., 'US', '150').
            decode: Optional human-readable decode for the code.
        """
        alias_code = None
        if code and decode:
            # Look up the standard code based on scope type
            std_code = (
                self._builder.iso3166_code(code)
                if type.code == "C25464"
                else self._builder.iso3166_region_code(code)
            )
            if std_code:
                alias_code = self._builder.alias_code(std_code)
            else:
                self._errors.error(
                    f"Failed to create standard code for '{code}', '{decode}'",
                    KlassMethodLocation(self.MODULE, "_create_scope"),
                )
        gs = self._builder.create(GeographicScope, {"type": type, "code": alias_code})
        if gs:
            self._errors.info(
                f"Created scope of type {type.decode}{f' -> [{code}, {decode}]' if code else ''}",
                KlassMethodLocation(self.MODULE, "_create_scope"),
            )
            results.append(gs)
        else:
            self._errors.error(
                f"Failed to create geographic scope for {type}, {code}, {decode}",
                KlassMethodLocation(self.MODULE, "_create_scope"),
            )



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00090(RuleTemplate):
    """
    DDF00090: The same Biomedical Concept Category must not be referenced more than once from the same activity.

    Applies to: Activity
    Attributes: bcCategories
    """

    def __init__(self):
        super().__init__(
            "DDF00090",
            RuleTemplate.ERROR,
            "The same Biomedical Concept Category must not be referenced more than once from the same activity.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0195(RuleTemplate):
    """
    CHK0195: Every identifier must be unique within the scope of an identified organization.

    Applies to: StudyIdentifier, ReferenceIdentifier, AdministrableProductIdentifier
    Attributes: text
    """

    def __init__(self):
        super().__init__(
            "CHK0195",
            RuleTemplate.ERROR,
            "Every identifier must be unique within the scope of an identified organization.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal
from .syntax_template import SyntaxTemplate


class Condition(SyntaxTemplate):
    contextIds: List[str] = []
    appliesToIds: List[str] = []
    instanceType: Literal["Condition"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00079(RuleTemplate):
    """
    DDF00079: If a synonym is specified then it is not expected to be equal to the name of the biomedical concept (case insensitive).

    Applies to: BiomedicalConcept
    Attributes: synonyms
    """

    def __init__(self):
        super().__init__(
            "DDF00079",
            RuleTemplate.ERROR,
            "If a synonym is specified then it is not expected to be equal to the name of the biomedical concept (case insensitive).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00117(RuleTemplate):
    """
    DDF00117: A study protocol document version's protocol status must be specified using the Protocol Status Value Set Terminology (C188723) DDF codelist.

    Applies to: StudyProtocolDocumentVersion
    Attributes: protocolStatus
    """

    def __init__(self):
        super().__init__(
            "DDF00117",
            RuleTemplate.ERROR,
            "A study protocol document version's protocol status must be specified using the Protocol Status Value Set Terminology (C188723) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0187(RuleTemplate):
    """
    CHK0187: A study definition document version must not be referenced more than once by the same study version.

    Applies to: StudyVersion
    Attributes: documentVersions
    """

    def __init__(self):
        super().__init__(
            "CHK0187",
            RuleTemplate.ERROR,
            "A study definition document version must not be referenced more than once by the same study version.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00076(RuleTemplate):
    """
    DDF00076: If a biomedical concept is referenced from an activity then it is not expected to be referenced as well by a biomedical concept category that is referenced from the same activity.

    Applies to: Activity, BiomedicalConceptCategory
    Attributes: biomedicalConcepts, members
    """

    def __init__(self):
        super().__init__(
            "DDF00076",
            RuleTemplate.ERROR,
            "If a biomedical concept is referenced from an activity then it is not expected to be referenced as well by a biomedical concept category that is referenced from the same activity.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00129(RuleTemplate):
    """
    DDF00129: A study intervention's product designation must be specified using the product designation (C207418) DDF codelist.

    Applies to: StudyIntervention
    Attributes: productDesignation
    """

    def __init__(self):
        super().__init__(
            "DDF00129",
            RuleTemplate.ERROR,
            "A study intervention's product designation must be specified using the product designation (C207418) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00111(RuleTemplate):
    """
    DDF00111: The unit of a planned age is expected to be specified using terms from the Age Unit (C66781) SDTM codelist.

    Applies to: StudyDesignPopulation, StudyCohort
    Attributes: plannedAge
    """

    def __init__(self):
        super().__init__(
            "DDF00111",
            RuleTemplate.ERROR,
            "The unit of a planned age is expected to be specified using terms from the Age Unit (C66781) SDTM codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .strength import Strength
from .code import Code


class Substance(ApiBaseModelWithIdNameLabelAndDesc):
    codes: List[Code] = []
    strengths: List[Strength] = []  # Not in API
    referenceSubstance: Union["Substance", None] = None
    instanceType: Literal["Substance"]



from typing import List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .comment_annotation import CommentAnnotation


class AnalysisPopulation(ApiBaseModelWithIdNameLabelAndDesc):
    text: str
    subsetOfIds: List[str] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["AnalysisPopulation"]



from usdm3.rules.library.rule_ddf00140 import RuleDDF00140 as V3Rule


class RuleDDF00140(V3Rule):
    def validate(self, config: dict) -> bool:
        return self._ct_check(config, "Organization", "type")



from typing import List, Union, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .quantity_range import Quantity
from .duration import Duration
from .alias_code import AliasCode
from .comment_annotation import CommentAnnotation


class Administration(ApiBaseModelWithIdNameLabelAndDesc):
    duration: Duration
    dose: Union[Quantity, None] = None
    route: Union[AliasCode, None] = None
    frequency: Union[AliasCode, None] = None
    administrableProductId: Union[str, None] = None
    medicalDeviceId: Union[str, None] = None
    notes: List[CommentAnnotation] = []
    instanceType: Literal["Administration"]



class BaseManager:
    def __init__(self):
        self._items = {}

    def __iter__(self):
        return iter(self._items)

    def clear(self) -> None:
        self._items = {}

    def add(self, name: str, value: str) -> None:
        self._items[name.upper()] = value

    def get(self, name: str) -> str:
        u_name = name.upper()
        return self._items[u_name] if u_name in self._items else ""

    def includes(self, name) -> bool:
        u_name = name.upper()
        return u_name in self._items



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0273(RuleTemplate):
    """
    CHK0273: A study design's study phase must be specified according to the extensible Trial Phase Response (C66737) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: ObservationalStudyDesign, InterventionalStudyDesign
    Attributes: studyPhase
    """

    def __init__(self):
        super().__init__(
            "CHK0273",
            RuleTemplate.ERROR,
            "A study design's study phase must be specified according to the extensible Trial Phase Response (C66737) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import enum
import datetime
from uuid import UUID
from pydantic import BaseModel


# Example, see https://stackoverflow.com/questions/10252010/serializing-class-instance-to-json
def serialize_as_json(obj: BaseModel):
    if isinstance(obj, enum.Enum):
        return obj.value
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    elif (
        hasattr(obj, "instanceType")
        and getattr(obj, "instanceType") == "ExtensionAttribute"
    ):
        obj_dict: dict = obj.__dict__
        return {k: v for k, v in obj_dict.items() if v is not None}
    else:
        return obj.__dict__



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00128(RuleTemplate):
    """
    DDF00128: A study intervention's type must be specified using the Intervention Type Response (C99078) SDTM codelist.

    Applies to: StudyIntervention
    Attributes: type
    """

    def __init__(self):
        super().__init__(
            "DDF00128",
            RuleTemplate.ERROR,
            "A study intervention's type must be specified using the Intervention Type Response (C99078) SDTM codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0120(RuleTemplate):
    """
    CHK0120: For a specified range at least a minimum or maximum value is expected.

    Applies to: Range
    Attributes: minValue, maxValue
    """

    def __init__(self):
        super().__init__(
            "CHK0120",
            RuleTemplate.ERROR,
            "For a specified range at least a minimum or maximum value is expected.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0254(RuleTemplate):
    """
    CHK0254: An administrable product's product designation must be specified using the product designation (C207418) DDF codelist.

    Applies to: StudyIntervention
    Attributes: productDesignation
    """

    def __init__(self):
        super().__init__(
            "CHK0254",
            RuleTemplate.ERROR,
            "An administrable product's product designation must be specified using the product designation (C207418) DDF codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0246(RuleTemplate):
    """
    CHK0246: The sponsor study role must point to exactly one organization.

    Applies to: StudyRole
    Attributes: organizations
    """

    def __init__(self):
        super().__init__(
            "CHK0246",
            RuleTemplate.ERROR,
            "The sponsor study role must point to exactly one organization.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

package_info = {}
with open("src/usdm4/__info__.py") as fp:
    exec(fp.read(), package_info)

setuptools.setup(
    name="usdm4",
    version=package_info["__package_version__"],
    author="D Iberson-Hurst",
    author_email="",
    description="A python package for using the CDISC TransCelerate USDM, version 4",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "usdm3==0.12.1",
        "simple_error_log>=0.7.0",
        "python-dateutil==2.9.0.post0",
    ],
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    package_data={
        "usdm4": [
            "ct/cdisc/library_cache/library_cache_all.yaml",
            "ct/cdisc/library_cache/library_cache_usdm.yaml",
            "ct/cdisc/config/ct_config.yaml",
            "ct/cdisc/missing/missing_ct.yaml",
            "ct/iso/iso3166/iso3166.json",
            "bc/cdisc/library_cache/library_cache.yaml",
            "rules/library/schema/usdm_v4-0-0.json",
        ]
    },
    tests_require=["pytest", "pytest-cov", "pytest-mock", "python-dotenv"],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
)



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .quantity_range import Quantity, Range


class Strength(ApiBaseModelWithIdNameLabelAndDesc):
    numerator: Union[Quantity, Range]
    denominator: Union[Quantity, None] = None
    instanceType: Literal["Strength"]



from typing import Literal, Union
from .api_base_model import ApiBaseModelWithId
from .quantity_range import Quantity, Range


class Duration(ApiBaseModelWithId):
    text: Union[str, None] = None
    quantity: Union[Quantity, Range, None] = None
    durationWillVary: bool
    reasonDurationWillVary: Union[str, None] = None
    instanceType: Literal["Duration"]



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00089(RuleTemplate):
    """
    DDF00089: Any parameter name referenced in a tag in the text should be specified in the data dictionary parameter maps.

    Applies to: EligibilityCriterion, Characteristic, Condition, Objective, Endpoint
    Attributes: text
    """

    def __init__(self):
        super().__init__(
            "DDF00089",
            RuleTemplate.ERROR,
            "Any parameter name referenced in a tag in the text should be specified in the data dictionary parameter maps.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from typing import Union, List, Literal
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .transition_rule import TransitionRule
from .comment_annotation import CommentAnnotation


class StudyElement(ApiBaseModelWithIdNameLabelAndDesc):
    transitionStartRule: Union[TransitionRule, None] = None
    transitionEndRule: Union[TransitionRule, None] = None
    studyInterventionIds: List[str] = []
    notes: List[CommentAnnotation] = []
    instanceType: Literal["StudyElement"]



from uuid import uuid4
from simple_error_log.errors import Errors
from simple_error_log.error_location import KlassMethodLocation
from usdm4.api.wrapper import Wrapper
from usdm4.api.code import Code
from usdm4.api.alias_code import AliasCode
from usdm4.api.geographic_scope import GeographicScope
from usdm4.api.governance_date import GovernanceDate
from usdm4.api.organization import Organization
from usdm4.api.study import Study
from usdm4.api.study_definition_document import StudyDefinitionDocument
from usdm4.api.study_definition_document_version import StudyDefinitionDocumentVersion
from usdm4.api.identifier import StudyIdentifier
from usdm4.api.study_title import StudyTitle
from usdm4.api.study_version import StudyVersion
from usdm4.api.biomedical_concept import BiomedicalConcept
from usdm4.api import __all__ as v4_classes
from usdm4.__info__ import __model_version__, __package_version__
from usdm3.base.id_manager import IdManager
from usdm3.base.api_instance import APIInstance
from usdm3.ct.cdisc.library import Library as CdiscCTLibrary
from usdm3.bc.cdisc.library import Library as CdiscBCLibrary
from usdm3.data_store.data_store import DataStore
from usdm4.ct.iso.iso3166.library import Library as Iso3166Library
from usdm4.ct.iso.iso639.library import Library as Iso639Library
from usdm4.builder.cross_reference import CrossReference
from usdm4.api.api_base_model import ApiBaseModelWithId
from usdm4.builder.other_ct_version_manager import OtherCTVersionManager


class Builder:
    MODULE = "usdm4.builder.builder.Builder"

    def __init__(self, root_path: str, errors: Errors):
        self._id_manager: IdManager = IdManager(v4_classes)
        self.errors = errors
        self.api_instance: APIInstance = APIInstance(self._id_manager)
        self.cdisc_ct_library = CdiscCTLibrary(root_path)
        self.cdisc_bc_library = CdiscBCLibrary(root_path, self.cdisc_ct_library)
        self.iso3166_library = Iso3166Library(root_path)
        self.iso639_library = Iso639Library(root_path)
        self.cross_reference = CrossReference()
        self.other_ct_version_manager = OtherCTVersionManager()
        self._data_store = None

        # Lazy loading: Track if CT libraries have been loaded
        self._ct_loaded = False
        self._cdisc_code_system = None
        self._cdisc_code_system_version = None

    def _ensure_ct_loaded(self):
        """Lazy load CT libraries only when first needed."""
        if not self._ct_loaded:
            self.cdisc_ct_library.load()
            self.cdisc_bc_library.load()
            self.iso3166_library.load()
            self.iso639_library.load()
            self._cdisc_code_system = self.cdisc_ct_library.system
            self._cdisc_code_system_version = self.cdisc_ct_library.version
            self._ct_loaded = True

    def clear(self):
        self._id_manager.clear()
        self.errors.clear()
        self.cross_reference.clear()
        self._data_store = None

    def seed(self, file_path: str):
        self._data_store = DataStore(file_path)
        self._data_store.decompose()
        for klass in v4_classes:
            for instance in self._data_store.instances_by_klass(klass):
                if "id" in instance:
                    self._id_manager.add_id(klass, instance["id"])

    def create(
        self, klass: str, params: dict, cross_reference: bool = True
    ) -> ApiBaseModelWithId | None:
        try:
            object: ApiBaseModelWithId = self.api_instance.create(klass, params)
            if self._check_object(object) and cross_reference:
                name = self._set_name(object, params)
                self.cross_reference.add(object, name)
            return object
        except Exception as e:
            # print(f"EXCEPTION ON CREATE: {e}")
            location = KlassMethodLocation(self.MODULE, "create")
            self.errors.exception(
                f"Failed to create instance of klass '{klass}' with params {params}, reason: {e}",
                e,
                location,
            )
            return None

    def _check_object(self, object) -> bool:
        return True if object and hasattr(object, "id") else False

    def _set_name(self, object, params: dict) -> str | None:
        if hasattr(object, "name"):
            return object.name
        elif "name" in params:
            return params["name"]
        return None

    def minimum(self, title: str, identifier: str, version: str) -> "Wrapper":
        """
        Create a minimum study with the given title, identifier, and version.
        """
        self._ensure_ct_loaded()
        # Define the codes to be used in the study
        english_code = self.iso639_code("en")
        title_type = self.cdisc_code("C207616", "Official Study Title")
        organization_type_code = self.cdisc_code("C54149", "Pharmaceutical Company")
        doc_status_code = self.cdisc_code("C25425", "Approved")
        protocol_code = self.cdisc_code("C70817", "Protocol")
        global_code = self.cdisc_code("C68846", "Global")
        global_scope = self.create(GeographicScope, {"type": global_code})
        approval_date_code = self.cdisc_code("C71476", "Approval Date")

        # Study Title
        study_title = self.create(StudyTitle, {"text": title, "type": title_type})

        # Governance dates
        approval_date = self.create(
            GovernanceDate,
            {
                "name": "D_APPROVE",
                "label": "Design Approval",
                "description": "Design approval date",
                "type": approval_date_code,
                "dateValue": "2006-06-01",
                "geographicScopes": [global_scope],
            },
        )
        # Define the organization and the study identifier
        organization = self.create(
            Organization,
            {
                "name": "Sponsor",
                "type": organization_type_code,
                "identifier": "To be provided",
                "identifierScheme": "To be provided",
                "legalAddress": None,
            },
        )
        study_identifier = self.create(
            StudyIdentifier,
            {"text": identifier, "scopeId": organization.id},
        )

        # Documenta
        study_definition_document_version = self.create(
            StudyDefinitionDocumentVersion,
            {"version": "1", "status": doc_status_code, "dateValues": [approval_date]},
        )
        study_definition_document = self.create(
            StudyDefinitionDocument,
            {
                "name": "PROTOCOL DOCUMENT",
                "label": "Protocol Document",
                "description": "The entire protocol document",
                "language": english_code,
                "type": protocol_code,
                "templateName": "Sponsor",
                "versions": [study_definition_document_version],
            },
        )

        study_version = self.create(
            StudyVersion,
            {
                "versionIdentifier": "1",
                "rationale": "To be provided",
                "titles": [study_title],
                "studyDesigns": [],
                "documentVersionIds": [study_definition_document_version.id],
                "studyIdentifiers": [study_identifier],
                "studyPhase": None,
                "dateValues": [approval_date],
                "amendments": [],
                "organizations": [organization],
            },
        )
        study = self.create(
            Study,
            {
                "id": str(uuid4()),
                "name": "Study",
                "label": title,
                "description": title,
                "versions": [study_version],
                "documentedBy": [study_definition_document],
            },
        )

        # Return the wrapper for the study
        result = self.api_instance.create(
            Wrapper,
            {
                "study": study,
                "usdmVersion": __model_version__,
                "systemName": "Python USDM4 Package",
                "systemVersion": __package_version__,
            },
        )
        return result

    def klass_and_attribute(self, klass: str, attribute: str) -> dict:
        self._ensure_ct_loaded()
        return self.cdisc_ct_library.klass_and_attribute(klass, attribute)

    def klass_and_attribute_value(self, klass: str, attribute: str, value: str) -> Code:
        self._ensure_ct_loaded()
        code_item, version = self.cdisc_ct_library.klass_and_attribute_value(
            klass, attribute, value
        )
        return (
            self.create(
                Code,
                {
                    "code": code_item["conceptId"],
                    "codeSystem": self._cdisc_code_system,
                    "codeSystemVersion": version,
                    "decode": code_item["preferredTerm"],
                },
            )
            if code_item
            else None
        )

    def cdisc_code(self, code: str, decode: str) -> Code:
        self._ensure_ct_loaded()
        cl = self.cdisc_ct_library.cl_by_term(code)
        return (
            self.create(
                Code,
                {
                    "code": code,
                    "codeSystem": self._cdisc_code_system,
                    "codeSystemVersion": cl["source"]["effective_date"],
                    "decode": decode,
                },
            )
            if cl
            else None
        )

    def cdisc_unit_code(self, unit: str) -> Code:
        self._ensure_ct_loaded()
        unit = self.cdisc_ct_library.unit(unit)
        unit_cl = self.cdisc_ct_library.unit_code_list()
        return (
            self.create(
                Code,
                {
                    "code": unit["conceptId"],
                    "codeSystem": self._cdisc_code_system,
                    "codeSystemVersion": unit_cl["source"]["effective_date"],
                    "decode": unit["preferredTerm"],
                },
            )
            if unit
            else None
        )

    def alias_code(self, standard_code: Code) -> AliasCode:
        return self.create(AliasCode, {"standardCode": standard_code})

    def bc(self, name) -> BiomedicalConcept | None:
        self._ensure_ct_loaded()
        if self.cdisc_bc_library.exists(name):
            bc_params = self.cdisc_bc_library.usdm(name)
            self._set_ids(bc_params)
            return self.create(
                BiomedicalConcept, bc_params, False
            )  # No cross reference for BCs
        else:
            return None

    def iso3166_code_or_decode(self, text: str) -> Code:
        self._ensure_ct_loaded()
        code, decode = self.iso3166_library.code_or_decode(text)
        if code:
            return self.create(
                Code,
                {
                    "code": code,
                    "codeSystem": self.iso3166_library.system,
                    "codeSystemVersion": self.iso3166_library.version,
                    "decode": decode,
                },
            )
        else:
            return None

    def iso3166_code(self, code: str) -> Code:
        self._ensure_ct_loaded()
        code, decode = self.iso3166_library.decode(code)
        if code:
            return self.create(
                Code,
                {
                    "code": code,
                    "codeSystem": self.iso3166_library.system,
                    "codeSystemVersion": self.iso3166_library.version,
                    "decode": decode,
                },
            )
        else:
            return None

    def iso639_code_or_decode(self, text: str) -> Code:
        self._ensure_ct_loaded()
        code, decode = self.iso639_library.code_or_decode(text)
        # print(f"ISO639: {text} = [{code} {decode}]")
        if code:
            return self.create(
                Code,
                {
                    "code": code,
                    "codeSystem": self.iso639_library.system,
                    "codeSystemVersion": self.iso639_library.version,
                    "decode": decode,
                },
            )
        else:
            return None

    def iso639_code(self, code: str) -> Code:
        self._ensure_ct_loaded()
        new_code, decode = self.iso639_library.decode(code)
        if new_code:
            return self.create(
                Code,
                {
                    "code": code,
                    "codeSystem": self.iso639_library.system,
                    "codeSystemVersion": self.iso639_library.version,
                    "decode": decode,
                },
            )
        else:
            return None

    def iso3166_region_code(self, code: str) -> Code:
        self._ensure_ct_loaded()
        code, decode = self.iso3166_library.region_code(code)
        return self.create(
            Code,
            {
                "code": code,
                "codeSystem": self.iso3166_library.system,
                "codeSystemVersion": self.iso3166_library.version,
                "decode": decode,
            },
        )

    def other_code(self, code: str, system: str, version: str, decode: str) -> Code:
        return self.create(
            Code,
            {
                "code": code,
                "codeSystem": system,
                "codeSystemVersion": version,
                "decode": decode,
            },
        )

    def sponsor(self, sponsor_name: str) -> Organization:
        self._ensure_ct_loaded()
        sponsor_code = self.cdisc_code("C54149", "Pharmaceutical Company")
        return self.create(
            Organization,
            {
                "name": sponsor_name,
                "label": sponsor_name,
                "type": sponsor_code,
                "identifier": "---------",
                "identifierScheme": "DUNS",
                # "legalAddress": address
            },
        )

    def double_link(self, items, prev_attribute, next_attribute):
        for idx, item in enumerate(items):
            if idx == 0:
                setattr(item, prev_attribute, None)
            else:
                the_id = getattr(items[idx - 1], "id")
                setattr(item, prev_attribute, the_id)
            if idx == len(items) - 1:
                setattr(item, next_attribute, None)
            else:
                the_id = getattr(items[idx + 1], "id")
                setattr(item, next_attribute, the_id)

    def load(self, data: dict):
        self._decompose(data)

    def _decompose(self, data) -> None:
        if isinstance(data, dict):
            self._add_id(data)
            for _, value in data.items():
                if isinstance(value, dict):
                    self._decompose(value)
                elif isinstance(value, list):
                    for item in value:
                        self._decompose(item)

    def _add_id(self, data: dict):
        self._id_manager.add_id(data["instanceType"], data["id"])

    def _set_ids(self, parent):
        if isinstance(parent, str) or isinstance(parent, bool) or (parent is None):
            return
        parent["id"] = self._id_manager.build_id(parent["instanceType"])
        for _, value in parent.items():
            if isinstance(value, list):
                for child in value:
                    self._set_ids(child)
            else:
                self._set_ids(value)



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00137(RuleTemplate):
    """
    DDF00137: References must be a fixed value or a reference to items stored elsewhere in the data model which must be specified in the correct format. They must start with '<usdm:ref', end with either '/>' or '></usdm:ref>', and must contain 'klass=\"klassName\"', 'id=\"idValue\"', and 'attribute=\"attributeName\"/>' in any order (where \"klassName\" and \"attributeName\" contain only letters in upper or lower case).

    Applies to: ParameterMap
    Attributes: reference
    """

    def __init__(self):
        super().__init__(
            "DDF00137",
            RuleTemplate.ERROR,
            "References must be a fixed value or a reference to items stored elsewhere in the data model which must be specified in the correct format. They must start with '<usdm:ref', end with either '/>' or '></usdm:ref>', and must contain 'klass=\"klassName\"', 'id=\"idValue\"', and 'attribute=\"attributeName\"/>' in any order (where \"klassName\" and \"attributeName\" contain only letters in upper or lower case).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



from usdm3.rules.library.rule_template import RuleTemplate


class RuleCHK0274(RuleTemplate):
    """
    CHK0274: A study design's study type must be specified using the Study Type Response (C99077) SDTM codelist.

    Applies to: ObservationalStudyDesign, InterventionalStudyDesign
    Attributes: studyType
    """

    def __init__(self):
        super().__init__(
            "CHK0274",
            RuleTemplate.ERROR,
            "A study design's study type must be specified using the Study Type Response (C99077) SDTM codelist.",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")
